from __future__ import print_function, division
from SLOPpy.subroutines.common import *
from SLOPpy.subroutines.spectral_subroutines import *
from SLOPpy.subroutines.io_subroutines import *
from SLOPpy.subroutines.fit_subroutines import *
from SLOPpy.subroutines.shortcuts import *
from SLOPpy.subroutines.rebin_subroutines import *
from SLOPpy.subroutines.clv_rm_subroutines import *
from SLOPpy.subroutines.math_functions import *
from astropy.convolution import convolve, Box1DKernel

__all__ = ['compute_spectra_lightcurve',
           'compute_spectra_lightcurve_clv_rm_correction',
           'plot_spectra_lightcurve',
           'plot_spectra_lightcurve_clv_rm_correction']


def compute_spectra_lightcurve_clv_rm_correction(config_in, lines_label):
    compute_spectra_lightcurve(config_in, lines_label)


def plot_spectra_lightcurve_clv_rm_correction(config_in, night_input=''):
    plot_spectra_lightcurve(config_in, night_input)

subroutine_name = 'spectra_lightcurve'
sampler_name = 'emcee'

def compute_spectra_lightcurve(config_in, lines_label):


    results_list_default = ['user',
                    'mcmc_night_MED',
                    'mcmc_night_MAP',
                    'mcmc_global_MED',
                    'mcmc_global_MAP']

    append_list = ['', '_uncorrected', '_clv_model']

    do_average_instead_of_sum = True

    night_dict = from_config_get_nights(config_in)
    #instrument_dict = from_config_get_instrument(config_in)
    #system_dict = from_config_get_system(config_in)
    planet_dict = from_config_get_planet(config_in)

    spectral_lines = from_config_get_spectral_lines(config_in)
    lines_dict = spectral_lines[lines_label]
    clv_rm_correction = lines_dict.get('clv_rm_correction', True)


    # from_config_get_transmission_lightcurve(config_in)
    #lightcurve_dict = from_config_get_transmission_lightcurve(config_in)

    shared_data = load_from_cpickle('shared', config_in['output'])

    """ Using the MCMC fit range to define the transmission spectrum region """

    shared_selection = (shared_data['coadd']['wave'] >= lines_dict['range'][0]) \
        & (shared_data['coadd']['wave'] < lines_dict['range'][1])

    preparation_template = {
        'subroutine': subroutine_name,
        'range': lines_dict['range'],
        'wave': shared_data['coadd']['wave'][shared_selection],
        'step': shared_data['coadd']['step'][shared_selection],
        'size': int(np.sum(shared_selection)),
    }

    if 'full_transit_duration' in planet_dict:
        full_transit_duration = planet_dict['total_transit_duration'][0]
    else:
        full_transit_duration = planet_dict['transit_duration'][0]

    if 'total_transit_duration' in planet_dict:
        total_transit_duration = planet_dict['total_transit_duration'][0]
    else:
        total_transit_duration = planet_dict['transit_duration'][0]

    """
        The transit phase [0-1] is divided in N (=5) bins. Two arrays are computed:
        - transit_in_bins: array with the boundaries of the bins, size=N+1
        - transit_in_step: average size of the bin, size=1
    """

    transit_in_bins = np.linspace(
        -total_transit_duration/2./planet_dict['period'][0],
        total_transit_duration/2./planet_dict['period'][0],
        6
    )
    transit_full_bins = np.linspace(
        -full_transit_duration/2./planet_dict['period'][0],
        full_transit_duration/2./planet_dict['period'][0],
        6
    )

    transit_in_step = np.average(transit_in_bins[1:]-transit_in_bins[:-1])
    transit_full_step = np.average(transit_full_bins[1:]-transit_full_bins[:-1])


    """ Preparation stage  - rebinning of spectra """

    for night in night_dict:

        preparation = None # Free up memory
        try:
            preparation = load_from_cpickle(subroutine_name + '_preparation', config_in['output'], night, lines_label)
            print("{0:45s} Night:{1:15s}   {2:s}".format(subroutine_name+ '_preparation', night, 'Retrieved'))
            continue
        except:
            print("{0:45s} Night:{1:15s}   {2:s}".format(subroutine_name+ '_preparation', night, 'Computing'))
            print()

        preparation = preparation_template.copy()

        """ Retrieving the list of observations"""
        lists = load_from_cpickle('lists', config_in['output'], night)

        calib_data = load_from_cpickle('calibration_fibA', config_in['output'], night)
        input_data = retrieve_observations( config_in['output'], night, lists['observations'])
        observational_pams = load_from_cpickle('observational_pams', config_in['output'], night)


        for n_obs, obs in enumerate( lists['observations']):

            preparation[obs] = {}

            preparation[obs]['rescaling'], \
            preparation[obs]['rescaled'], \
            preparation[obs]['rescaled_err'] = perform_rescaling(
                input_data[obs]['wave'], input_data[obs]['e2ds'], input_data[obs]['e2ds_err'],
                observational_pams['wavelength_rescaling'])

            preserve_flux = input_data[obs].get('absolute_flux', True)

            preparation[obs]['rebinned'] = \
                rebin_2d_to_1d(input_data[obs]['wave'],
                               input_data[obs]['step'],
                               input_data[obs]['e2ds'],
                               calib_data['blaze'],
                               preparation['wave'],
                               preparation['step'],
                               preserve_flux=preserve_flux,
                               rv_shift=observational_pams[obs]['rv_shift_ORF2SRF'])

            preparation[obs]['rebinned_err'] = \
                rebin_2d_to_1d(input_data[obs]['wave'],
                               input_data[obs]['step'],
                               input_data[obs]['e2ds_err'],
                               calib_data['blaze'],
                               preparation['wave'],
                               preparation['step'],
                               preserve_flux=preserve_flux,
                               rv_shift=observational_pams[obs]['rv_shift_ORF2SRF'])


        save_to_cpickle(subroutine_name+'_preparation', preparation, config_in['output'], night, lines_label)

        # Free up memory
        calib_data = None
        input_data = None
        observational_pams = None

    """ Actual computation of spectral lightcurve """
    # doublet sodium in the lab reference frame

    """
    C stands for central
    """
    C_bands = {}
    for passband_key, passband_val in lines_dict['passbands'].items():
        C_bands[passband_key] = {}
        for line_key, line_val in lines_dict['lines'].items():
            C_bands[passband_key][line_key] = (np.abs(preparation['wave'] - line_val) < passband_val / 2.)

    """
    S stands for side
    """
    S_bands = {}
    for band_key, band_val in lines_dict['continuum'].items():
        S_bands[band_key] = (preparation['wave'] >= band_val[0]) & (preparation['wave'] <= band_val[1])



    results_list = results_list_default.copy()

    for results_selection in results_list_default:
        skip_iteration = False

        for night in night_dict:

            print_warning = True
            if skip_iteration: continue

            
            binned_mcmc_night = check_existence_cpickle(
                    'transmission_binned_mcmc_'+sampler_name+'_results', config_in['output'], night, lines_label)
            binned_mcmc_global = check_existence_cpickle(
                    'transmission_binned_mcmc_'+sampler_name+'_results', config_in['output'], lines_label)

            mcmc_night = check_existence_cpickle(
                    'transmission_mcmc_'+sampler_name+'_results', config_in['output'], night, lines_label)
            mcmc_global = check_existence_cpickle(
                    'transmission_mcmc_'+sampler_name+'_results', config_in['output'], lines_label)

            results_list = ['user']
            if (mcmc_night or binned_mcmc_night):
                results_list.append(['mcmc_night_MED', 'mcmc_night_MAP'])
            if (mcmc_global or binned_mcmc_global):
                results_list.append(['mcmc_global_MED', 'mcmc_global_MAP'])
            if results_selection not in results_list:
                print('  {0:s} results not found, skipping iteration'.format(results_selection))
                skip_iteration = True
                continue


            if mcmc_night and results_selection in  ['mcmc_night_MED', 'mcmc_night_MAP']:
                mcmc_results_night = load_from_cpickle(
                    'transmission_mcmc_'+sampler_name+'_results', config_in['output'], night, lines_label)

                print('  Observational parameters from MCMC fit of unbinned data, individual night')

            elif mcmc_global and results_selection in  ['mcmc_global_MED', 'mcmc_global_MAP']:
                mcmc_results_global = load_from_cpickle(
                    'transmission_mcmc_'+sampler_name+'_results', config_in['output'], lines=lines_label)

                print('  Observational parameters from MCMC fit of unbinned data, global fit')

            elif binned_mcmc_night and results_selection in  ['mcmc_night_MED', 'mcmc_night_MAP']:
                mcmc_results_night = load_from_cpickle(
                    'transmission_binned_mcmc_'+sampler_name+'_results', config_in['output'], night, lines_label)

                print('  Observational parameters from MCMC fit of binned data, individual night')

            elif binned_mcmc_global and results_selection in  ['mcmc_global_MED', 'mcmc_global_MAP']:
                mcmc_results_global = load_from_cpickle(
                    'transmission_binned_mcmc_'+sampler_name+'_results', config_in['output'], lines=lines_label)

                print('  Observational parameters from MCMC fit of binned data, global fit')
            else:
                print('  Observational parameters from configuration file')

            try:
                lightcurve = load_from_cpickle(subroutine_name + '_' + results_selection , config_in['output'], night, lines_label)
                print("{0:45s} Night:{1:15s}   {2:s}".format(subroutine_name, night, 'Retrieved'))
                continue
            except:
                print("{0:45s} Night:{1:15s}   {2:s}".format(subroutine_name, night, 'Computing'))
                print()

            preparation = load_from_cpickle(subroutine_name + '_preparation', config_in['output'], night, lines_label)

            """ Retrieving the list of observations"""
            lists = load_from_cpickle('lists', config_in['output'], night)
            observational_pams = load_from_cpickle('observational_pams', config_in['output'], night)

            processed = {
                'subroutine': 'compute_spectra_lightcurve',
                'range': preparation['range'],
                'wave': preparation['wave'],
                'step': preparation['step'],
                'size': preparation['size']
            }

            lightcurve = {
                'subroutine': subroutine_name,
                'arrays': {
                    'observations': {
                        'obs_name': np.zeros(len(lists['observations']), dtype=str),
                        'phase': np.zeros(len(lists['observations'])),
                    },
                    'transit_in': {},
                    'transit_full': {},
                    'transit_out': {},
                },
                'C_bands': C_bands,
                'S_bands': S_bands,
                'average': {},
                'bins': {
                    'transit_in_bins': transit_in_bins,
                    'transit_in_step': transit_in_step,
                    'transit_full_bins': transit_full_bins,
                    'transit_full_step': transit_full_step
                }
            }
            """ Adding the C-bands arrays to the dictionary"""

            for band_key in C_bands:
                for name_append in append_list:
                    lightcurve['arrays']['observations']['ratio_' + band_key + name_append] = np.zeros([len(lists['observations']), 2])

            transit_out_flag = np.zeros(len(lists['observations']), dtype=bool)
            transit_in_flag = np.zeros(len(lists['observations']), dtype=bool)
            transit_full_flag = np.zeros(len(lists['observations']), dtype=bool)

            if clv_rm_correction:
                try:
                    clv_rm_models = load_from_cpickle('clv_rm_models', config_in['output'], night, lines_label)
                except (FileNotFoundError, IOError):
                    clv_rm_models = load_from_cpickle('clv_rm_models', config_in['output'], night)



            """ Shift into planetary reference system is the default
            choice"""
            if results_selection == 'user':
                planet_R_factor = observational_pams.get('Rp_factor', 1.00000)
            elif results_selection == 'mcmc_night_MED':
                planet_R_factor = mcmc_results_night['results']['planet_R']
            elif results_selection == 'mcmc_night_MAP':
                planet_R_factor = mcmc_results_night['results_MAP']['planet_R']
            elif results_selection == 'mcmc_global_MED':
                planet_R_factor = mcmc_results_global['results']['planet_R']
            elif results_selection == 'mcmc_global_MAP':
                planet_R_factor = mcmc_results_global['results_MAP']['planet_R']

            for n_obs, obs in enumerate( lists['observations']):

                processed[obs] = {}
                lightcurve[obs] = {}


                processed[obs]['uncorrected'] = preparation[obs]['rebinned']
                processed[obs]['uncorrected_err'] = preparation[obs]['rebinned_err']

                if clv_rm_correction:

                    """" CLV + RM computation in the planetary reference frame """
                    processed[obs]['clv_model_stellarRF'] = interpolate1d_grid_nocheck(planet_R_factor,
                                                                                        clv_rm_models['common']['radius_grid'],
                                                                                        clv_rm_models[obs]['clv_rm_model_convolved_normalized'])

                    processed[obs]['clv_model_rebinned'] = \
                        rebin_1d_to_1d(clv_rm_models['common']['wave'],
                                        clv_rm_models['common']['step'],
                                        processed[obs]['clv_model_stellarRF'],
                                        processed['wave'],
                                        processed['step'],
                                        preserve_flux=False)

                    processed[obs]['rebinned'] = processed[obs]['uncorrected'] / processed[obs]['clv_model_rebinned']
                    processed[obs]['rebinned_err'] = processed[obs]['uncorrected_err'] / processed[obs]['clv_model_rebinned']

                else:
                    processed[obs]['clv_model_rebinned'] = np.ones(processed['size'])
                    processed[obs]['rebinned'] = processed[obs]['uncorrected']
                    processed[obs]['rebinned_err'] = processed[obs]['uncorrected_err']
                    if print_warning:
                        print('   *** No CLV correction')
                        print_warning = False


                try:
                    phase_internal = (observational_pams[obs]['BJD'] - night_dict[night]['time_of_transit'][0])/planet_dict['period'][0]
                except:
                    phase_internal = (observational_pams[obs]['BJD'] - night_dict[night]['time_of_transit'])/planet_dict['period'][0]


                processed[obs]['bands'] = {
                    'phase': phase_internal
                }
                processed[obs]['bands_uncorrected'] = {
                    'phase': phase_internal
                }
                processed[obs]['bands_clv_model'] = {
                    'phase': phase_internal
                }

                processed[obs]['s_integrated'] = 0.000
                processed[obs]['s_integrated_uncorrected'] = 0.000
                processed[obs]['s_integrated_clv_model'] = 0.000
                processed[obs]['s_sigmaq_sum'] = 0.000

                n_bands = 0.00
                for band_key, band_val in S_bands.items():
                    if do_average_instead_of_sum:
                        processed[obs]['bands'][band_key] =  \
                            [np.average(processed[obs]['rebinned'][band_val]),
                            np.sum((processed[obs]['rebinned_err'][band_val])**2)
                            /len(processed[obs]['rebinned_err'][band_val])**2]
                        processed[obs]['bands_uncorrected'][band_key] =  \
                            [np.average(processed[obs]['uncorrected'][band_val]),
                            np.sum((processed[obs]['uncorrected_err'][band_val])**2)
                            /len(processed[obs]['uncorrected_err'][band_val])**2]
                        processed[obs]['bands_clv_model'][band_key] =  \
                            [np.average(processed[obs]['clv_model_rebinned'][band_val]),
                            np.sum((processed[obs]['rebinned_err'][band_val])**2)
                            /len(processed[obs]['rebinned_err'][band_val])**2]
                    else:
                        processed[obs]['bands'][band_key] =  \
                            [np.sum(processed[obs]['rebinned'][band_val]),
                            np.sum((processed[obs]['rebinned_err'][band_val])**2)]
                        processed[obs]['bands_uncorrected'][band_key] =  \
                            [np.sum(processed[obs]['uncorrected'][band_val]),
                            np.sum((processed[obs]['uncorrected_err'][band_val])**2)]
                        processed[obs]['bands_clv_model'][band_key] =  \
                            [np.sum(processed[obs]['clv_model_rebinned'][band_val]),
                            np.sum((processed[obs]['rebinned_err'][band_val])**2)]

                    processed[obs]['s_integrated'] += processed[obs]['bands'][band_key][0]
                    processed[obs]['s_integrated_uncorrected'] += processed[obs]['bands_uncorrected'][band_key][0]
                    processed[obs]['s_integrated_clv_model'] += processed[obs]['bands_clv_model'][band_key][0]
                    processed[obs]['s_sigmaq_sum'] += processed[obs]['bands'][band_key][1]
                    n_bands += 1.

                #todo:  why a 2 denominator??? 
                processed[obs]['s_integrated'] /= (n_bands / 2.)
                processed[obs]['s_integrated_uncorrected'] /= (n_bands / 2.)
                processed[obs]['s_integrated_clv_model'] /= (n_bands / 2.)
                processed[obs]['s_sigmaq_sum'] /= (n_bands / 2.)**2


                for band_key, band_dict in C_bands.items():
                    processed[obs]['bands'][band_key] = {}
                    processed[obs]['bands_uncorrected'][band_key] = {}
                    processed[obs]['bands_clv_model'][band_key] = {}

                    processed[obs]['c_integrated'] = 0.000
                    processed[obs]['c_integrated_uncorrected'] = 0.000
                    processed[obs]['c_integrated_clv_model'] = 0.000
                    processed[obs]['c_sigmaq_sum'] = 0.000

                    n_bands = 0.00

                    for line_key, line_val in band_dict.items():
                        if do_average_instead_of_sum:
                            processed[obs]['bands'][band_key][line_key] = \
                                [np.average(processed[obs]['rebinned'][line_val]),
                                np.sum((processed[obs]['rebinned_err'][line_val]) ** 2)
                                / len(processed[obs]['rebinned_err'][line_val]) ** 2]
                            processed[obs]['bands_uncorrected'][band_key][line_key] = \
                                [np.average(processed[obs]['uncorrected'][line_val]),
                                np.sum((processed[obs]['uncorrected_err'][line_val]) ** 2)
                                / len(processed[obs]['rebinned_err'][line_val]) ** 2]
                            processed[obs]['bands_clv_model'][band_key][line_key] = \
                                [np.average(processed[obs]['clv_model_rebinned'][line_val]),
                                np.sum((processed[obs]['rebinned_err'][line_val]) ** 2)
                                / len(processed[obs]['rebinned_err'][line_val]) ** 2]
                        else:
                            processed[obs]['bands'][band_key][line_key] = \
                                [np.sum(processed[obs]['rebinned'][line_val]),
                                np.sum((processed[obs]['rebinned_err'][line_val]) ** 2)]

                        processed[obs]['c_integrated'] += processed[obs]['bands'][band_key][line_key][0]
                        processed[obs]['c_integrated_uncorrected'] += processed[obs]['bands_uncorrected'][band_key][line_key][0]
                        processed[obs]['c_integrated_clv_model'] += processed[obs]['bands_clv_model'][band_key][line_key][0]
                        processed[obs]['c_sigmaq_sum'] += processed[obs]['bands'][band_key][line_key][1]
                        n_bands += 1.



                    processed[obs]['c_integrated'] /= (n_bands / 2.)
                    processed[obs]['c_integrated_uncorrected'] /= (n_bands / 2.)
                    processed[obs]['c_integrated_clv_model'] /= (n_bands / 2.)
                    processed[obs]['c_sigmaq_sum'] /= (n_bands / 2.) ** 2


                    for name_append in append_list:

                        ratio = processed[obs]['c_integrated' + name_append] / processed[obs]['s_integrated' + name_append]
                        ratio_err = ratio * np.sqrt(
                            processed[obs]['c_sigmaq_sum'] / processed[obs]['c_integrated' + name_append] ** 2
                            + processed[obs]['s_sigmaq_sum'] /  processed[obs]['s_integrated' + name_append] ** 2)

                        lightcurve[obs]['ratio_' + band_key + name_append] = [ratio, ratio_err]

                        lightcurve['arrays']['observations']['ratio_' + band_key + name_append][n_obs, :] = \
                            lightcurve[obs]['ratio_' + band_key + name_append][:]

                lightcurve[obs]['phase'] = processed[obs]['bands']['phase']
                lightcurve['arrays']['observations']['obs_name'][n_obs] = obs
                lightcurve['arrays']['observations']['phase'][n_obs] = lightcurve[obs]['phase']

                if obs in lists['transit_out']:
                    transit_out_flag[n_obs] = True
                if obs in lists['transit_in']:
                    transit_in_flag[n_obs] = True
                if obs in lists['transit_full']:
                    transit_full_flag[n_obs] = True

            for band_key in C_bands:
                for name_append in append_list:
                    lightcurve['arrays']['rescaling_' + band_key + name_append] = \
                        np.average(lightcurve['arrays']['observations']['ratio_' + band_key + name_append][transit_out_flag, 0], axis=0)

            sorting_index = np.argsort(lightcurve['arrays']['observations']['phase'])

            transit_out_flag = transit_out_flag[sorting_index]
            transit_in_flag = transit_in_flag[sorting_index]
            transit_full_flag = transit_full_flag[sorting_index]

            lightcurve['arrays']['observations']['obs_name'] = lightcurve['arrays']['observations']['obs_name'][sorting_index]
            lightcurve['arrays']['observations']['phase'] = lightcurve['arrays']['observations']['phase'][sorting_index]

            lightcurve['arrays']['transit_in']['obs_name'] = lightcurve['arrays']['observations']['obs_name'][transit_in_flag]
            lightcurve['arrays']['transit_in']['phase'] = lightcurve['arrays']['observations']['phase'][transit_in_flag]

            lightcurve['arrays']['transit_full']['obs_name'] = lightcurve['arrays']['observations']['obs_name'][transit_full_flag]
            lightcurve['arrays']['transit_full']['phase'] = lightcurve['arrays']['observations']['phase'][transit_full_flag]

            lightcurve['arrays']['transit_out']['obs_name'] = lightcurve['arrays']['observations']['obs_name'][transit_out_flag]
            lightcurve['arrays']['transit_out']['phase'] = lightcurve['arrays']['observations']['phase'][transit_out_flag]

            for band_key in C_bands:
                for name_append in append_list:

                    lightcurve['arrays']['observations']['ratio_' + band_key + name_append] = \
                        lightcurve['arrays']['observations']['ratio_' + band_key + name_append][sorting_index] \
                        / lightcurve['arrays']['rescaling_' + band_key + name_append]
                    lightcurve['arrays']['transit_in']['ratio_' + band_key + name_append] = \
                        lightcurve['arrays']['observations']['ratio_' + band_key + name_append][transit_in_flag]
                    lightcurve['arrays']['transit_full']['ratio_' + band_key + name_append] = \
                        lightcurve['arrays']['observations']['ratio_' + band_key + name_append][transit_full_flag]
                    lightcurve['arrays']['transit_out']['ratio_' + band_key + name_append] = \
                        lightcurve['arrays']['observations']['ratio_' + band_key + name_append][transit_out_flag]

                    avg_out, avg_out_sq = \
                        np.average(lightcurve['arrays']['transit_out']['ratio_' + band_key + name_append][:, 0],
                                weights=1./(lightcurve['arrays']['transit_out']['ratio_' + band_key + name_append][:, 1])**2,
                                returned=True)
                    avg_in, avg_in_sq = \
                        np.average(lightcurve['arrays']['transit_in']['ratio_' + band_key + name_append][:, 0],
                                weights=1. / (lightcurve['arrays']['transit_in']['ratio_' + band_key + name_append][:, 1]) ** 2,
                                returned=True)
                    avg_full, avg_full_sq = \
                        np.average(lightcurve['arrays']['transit_full']['ratio_' + band_key + name_append][:, 0],
                                weights=1. / (lightcurve['arrays']['transit_full']['ratio_' + band_key + name_append][:, 1]) ** 2,
                                returned=True)

                    lightcurve['average'][band_key + name_append] = {
                        'average_out': np.asarray([avg_out, 1./np.power(avg_out_sq, 0.5)]),
                        'average_in': np.asarray([avg_in, 1. / np.power(avg_in_sq, 0.5)]),
                        'average_full': np.asarray([avg_full, 1. / np.power(avg_full_sq, 0.5)]),
                    }

                    delta_fac = (lightcurve['average'][band_key + name_append]['average_full'][0]
                                 / lightcurve['average'][band_key + name_append]['average_out'][0])

                    delta_err = delta_fac * np.sqrt(
                        (lightcurve['average'][band_key + name_append]['average_out'][1]
                        / lightcurve['average'][band_key + name_append]['average_out'][0]) ** 2
                        + (lightcurve['average'][band_key + name_append]['average_full'][1]
                        / lightcurve['average'][band_key + name_append]['average_full'][0]) ** 2)

                    lightcurve['average'][band_key + name_append]['delta'] = np.asarray([(1.-delta_fac)*100., delta_err*100.])

            lightcurve['arrays']['observations']['transit_out_flag'] = transit_out_flag
            lightcurve['arrays']['observations']['transit_in_flag'] = transit_in_flag
            lightcurve['arrays']['observations']['transit_full_flag'] = transit_full_flag

            """ Compute the duration of the pre-transit observations, using as scale
                the number of bins, with the same size as those used inside the
                transit.

                The value is given by the difference of the phase of the beginning of the transit minus
                the phase of the first observation, keeping in mind that the centre of the transit has phase = 0

                An additional bin is added if there are observations left out from the actual number of bins
            """
            pre_duration = transit_full_bins[0] - lightcurve['arrays']['transit_out']['phase'][0]
            if pre_duration > 0:
                nsteps_pre = int(pre_duration/transit_full_step)
                if pre_duration % transit_full_step > 0.0:
                    nsteps_pre += 1
            else:
                nsteps_pre = 0

            """ same as pre-transit, but suing the post-transit instead"""
            post_duration = lightcurve['arrays']['transit_out']['phase'][-1] - transit_full_bins[-1]
            if post_duration > 0:
                nsteps_post = int(post_duration / transit_full_step)
                if post_duration % transit_full_step > 0.0:
                    nsteps_post += 1
            else:
                nsteps_post = 0

            """ THe full array with both in-transit and out-transit phase, built in such a way that the 
                - the lower boundary of the first in-transit bin corresponds to the beginning of the transit
                - the upper boundary of the last in-transit bin corresponds to the end of the transit
            """
            transit_bins = np.arange(transit_full_bins[0]-nsteps_pre*transit_full_step,
                                    transit_full_bins[-1] + (nsteps_post+1.1) * transit_full_step,
                                    transit_full_step)

            lightcurve['binned'] = {
                'observations': {
                    'phase': np.zeros(len(transit_bins)),
                },
                'transit_in': {},
                'transit_full': {},
                'transit_out': {},
            }
            for band_key in C_bands:
                for name_append in append_list:
                    lightcurve['binned']['observations']['ratio_' + band_key + name_append] = np.zeros([len(transit_bins), 2])

            transit_out_flag = np.zeros(len(transit_bins), dtype=bool)
            transit_in_flag = np.zeros(len(transit_bins), dtype=bool)
            transit_full_flag = np.zeros(len(transit_bins), dtype=bool)

            n_a = 0
            for nb in range(0, len(transit_bins)-1):
                sel = (lightcurve['arrays']['observations']['phase'] >= transit_bins[nb]) \
                    & (lightcurve['arrays']['observations']['phase'] < transit_bins[nb+1])

                if np.sum(sel) <= 0: continue
                lightcurve['binned']['observations']['phase'][n_a] = np.average(lightcurve['arrays']['observations']['phase'][sel])

                for band_key in C_bands:
                    for name_append in append_list:

                        lightcurve['binned']['observations']['ratio_' + band_key + name_append][n_a, 0], sum_weights = np.average(
                            lightcurve['arrays']['observations']['ratio_' + band_key + name_append][sel, 0],
                            weights=1. / lightcurve['arrays']['observations']['ratio_' + band_key + name_append][sel, 1]**2,
                            returned=True)

                        lightcurve['binned']['observations']['ratio_' + band_key + name_append][n_a, 1] = np.sqrt(1. / sum_weights)

                if np.abs(lightcurve['binned']['observations']['phase'][n_a]) >= \
                    total_transit_duration/2./planet_dict['period'][0]:
                        transit_out_flag[n_a] = True
                elif np.abs(lightcurve['binned']['observations']['phase'][n_a]) >= \
                    full_transit_duration/2./planet_dict['period'][0]:
                        transit_in_flag[n_a] = True
                else:
                    transit_full_flag[n_a] = True

                n_a += 1 # bins actually computed

            lightcurve['binned']['transit_in']['phase'] = lightcurve['binned']['observations']['phase'][transit_in_flag]
            lightcurve['binned']['transit_full']['phase'] = lightcurve['binned']['observations']['phase'][transit_full_flag]
            lightcurve['binned']['transit_out']['phase'] = lightcurve['binned']['observations']['phase'][transit_out_flag]
            lightcurve['binned']['observations']['phase'] = lightcurve['binned']['observations']['phase'][:n_a]

            for band_key in C_bands:
                for name_append in append_list:
                    lightcurve['binned']['transit_in']['ratio_' + band_key + name_append] = \
                        lightcurve['binned']['observations']['ratio_' + band_key + name_append][transit_in_flag, :]
                    lightcurve['binned']['transit_full']['ratio_' + band_key + name_append] = \
                        lightcurve['binned']['observations']['ratio_' + band_key + name_append][transit_full_flag, :]
                    lightcurve['binned']['transit_out']['ratio_' + band_key + name_append] = \
                        lightcurve['binned']['observations']['ratio_' + band_key + name_append][transit_out_flag, :]
                    lightcurve['binned']['observations']['ratio_' + band_key + name_append] = \
                        lightcurve['binned']['observations']['ratio_' + band_key + name_append][:n_a, :]

            save_to_cpickle(subroutine_name+ '_' + results_selection + '_processed', processed, config_in['output'], night, lines_label)
            save_to_cpickle(subroutine_name+ '_' + results_selection, lightcurve, config_in['output'], night, lines_label)

            # Forcing memory deallocation
            lightcurve = None
            processed = None
            preparation = None
            clv_rm_models = None

def plot_spectra_lightcurve(config_in, night_input='', clv_rm_correction=False):
    import matplotlib.pyplot as plt

    if clv_rm_correction:
        subroutine_name = 'spectra_lightcurve_clv_rm_correction'
    else:
        subroutine_name = 'spectra_lightcurve'

    night_dict = from_config_get_nights(config_in)

    if night_input=='':
        night_list = night_dict
    else:
        night_list = np.atleast_1d(night_input)

    for night in night_list:

        """ Retrieving the analysis"""
        try:
            lightcurve = load_from_cpickle(subroutine_name, config_in['output'], night)
            print("{0:45s} Night:{1:15s}   {2:s}".format(subroutine_name, night, 'Plotting'))
        except:
            print("{0:45s} Night:{1:15s}   {2:s}".format(subroutine_name, night, 'Skipped'))
            continue

        #observational_pams = load_from_cpickle('observational_pams', config_in['output'], night)
        C_bands = lightcurve['C_bands']


        print()
        for band_key in C_bands:
            print("Night: {0:s}   Band: {1:s}   Delta:{2:8.4f} +- {3:8.4f} [%]".format(night, band_key,
                  lightcurve['average'][band_key]['delta'][0], lightcurve['average'][band_key]['delta'][1]))

        for band_key in C_bands:


            plt.figure(figsize=(12, 6))
            plt.title('Spectra lightcurve - night {0:s} \n {1:s}'.format(night, band_key))
            plt.errorbar(lightcurve['arrays']['observations']['phase'],
                        lightcurve['arrays']['observations']['ratio_' + band_key][:,0]*100 -100.,
                         yerr= lightcurve['arrays']['observations']['ratio_' + band_key][:,1]*100 ,
                         fmt='.', c='k', alpha=0.25, label='observations')
            plt.errorbar(lightcurve['binned']['observations']['phase'],
                        lightcurve['binned']['observations']['ratio_' + band_key][:, 0]*100 -100.,
                         yerr= lightcurve['binned']['observations']['ratio_' + band_key][:,1]*100 ,
                         fmt='.', c='k', alpha=1.0, label='observations')

            plt.axvspan(-1, lightcurve['bins']['transit_in_bins'][0], alpha=0.25, color='green')
            plt.axvspan(lightcurve['bins']['transit_in_bins'][-1], 1., alpha=0.25, color='green')

            plt.axhline(0, c='C1')
            plt.xlim(lightcurve['arrays']['observations']['phase'][0]-0.01,
                     lightcurve['arrays']['observations']['phase'][-1]+0.01)
            plt.xlabel('orbital phase')
            plt.ylabel('$\mathcal{R}$ - 1. [%]')
            plt.legend()
            plt.show()

    print()



from __future__ import print_function, division
from SLOPpy.subroutines.common import *

def make_color_array(lists, input_data):
    """ Creation of the color array, based on the BJD of the observations
    """
    bjd = []
    am = []

    for obs in lists['observations']:
        bjd.append(input_data[obs]['BJD'] - 2450000.0)
        am.append(input_data[obs]['AIRMASS'])

    colors = np.asarray(bjd)
    cmap = plt.cm.viridis
    #cmap = plt.cm.Spectral
    line_colors = cmap(np.linspace(0, 1, len(lists['observations'])))

    return colors, cmap, line_colors

def make_color_array_matplotlib3(lists, input_data):
    """ Creation of the color array, based on the BJD of the observations
    """
    bjd = []
    mbjd = []
    am = []


    for obs in lists['observations']:
        bjd.append(input_data[obs]['BJD'])
        mbjd.append(input_data[obs]['mBJD'])
        am.append(input_data[obs]['AIRMASS'])

    color_cmap = plt.cm.viridis
    color_bjd_norm = plt.Normalize(vmin=bjd[0], vmax=bjd[-1])

    colors_bjd = color_cmap(color_bjd_norm(np.asarray(bjd)))

    color_am_norm = plt.Normalize(vmin=np.amin(am), vmax=np.amax(am))
    colors_am = color_cmap(color_am_norm(np.asarray(am)))

    colors_plot = {
        'BJD' : {},
        'mBJD' : {},
        'AIRMASS' : {}
    }
    colors_scatter = {
        'BJD' : {},
        'mBJD' : {},
        'AIRMASS' : {}
    }
    colors_properties = {
        'norm' : {
            'BJD': plt.Normalize(vmin=bjd[0], vmax=bjd[-1]),
            'mBJD': plt.Normalize(vmin=mbjd[0], vmax=mbjd[-1]),
            'AIRMASS': plt.Normalize(vmin=np.amin(am), vmax=np.amax(am))
        },
        'cmap' : plt.cm.viridis
    }

    for obs in lists['observations']:
        colors_plot['mBJD'][obs] = colors_properties['cmap'](
            colors_properties['norm']['mBJD'](input_data[obs]['mBJD']))
        colors_plot['BJD'][obs] = colors_properties['cmap'](
            colors_properties['norm']['BJD'](input_data[obs]['BJD']))
        colors_plot['AIRMASS'][obs] = colors_properties['cmap'](
            colors_properties['norm']['AIRMASS'](input_data[obs]['AIRMASS']))

        colors_scatter['mBJD'][obs] = [colors_properties['cmap'](
            colors_properties['norm']['mBJD'](input_data[obs]['mBJD']))[:-1]]
        colors_scatter['BJD'][obs] = [colors_properties['cmap'](
            colors_properties['norm']['BJD'](input_data[obs]['BJD']))[:-1]]
        colors_scatter['AIRMASS'][obs] = [colors_properties['cmap'](
            colors_properties['norm']['AIRMASS'](input_data[obs]['AIRMASS']))[:-1]]

    return colors_properties, colors_plot, colors_scatter

def grid_1plot():
    fig = plt.figure(figsize=(12, 6))
    gs = GridSpec(1, 2, width_ratios=[50, 1])

    ax = plt.subplot(gs[0, 0])
    cbax1 = plt.subplot(gs[0, 1])

    fig.subplots_adjust(wspace=0.04, hspace=0.25)

    return fig, gs, cbax1, ax

def grid_2plot(sharex=True, sharey=True):
    fig = plt.figure(figsize=(12, 6))
    gs = GridSpec(2, 2, width_ratios=[50, 1])

    ax1 = plt.subplot(gs[0, 0])

    if sharey and sharex:
        ax2 = plt.subplot(gs[1, 0], sharex=ax1, sharey=ax1)
    elif sharex:
        ax2 = plt.subplot(gs[1, 0], sharex=ax1)
    elif sharey:
        ax2 = plt.subplot(gs[1, 0], sharey=ax1)
    else:
        ax2 = plt.subplot(gs[1, 0])

    cbax1 = plt.subplot(gs[:, 1])

    fig.subplots_adjust(wspace=0.04, hspace=0.25)

    return fig, gs, cbax1, ax1, ax2

def grid_3plot_small(sharex=False, sharey=False, partial_share=True):
    fig = plt.figure(figsize=(12, 9))
    gs = GridSpec(3, 2, width_ratios=[50, 1], height_ratios = [3, 1, 3])

    ax1 = plt.subplot(gs[0, 0])

    if sharey and sharex:
        ax2 = plt.subplot(gs[1, 0], sharex=ax1, sharey=ax1)
        ax3 = plt.subplot(gs[2, 0], sharex=ax1, sharey=ax1)
    elif sharex:
        ax2 = plt.subplot(gs[1, 0], sharex=ax1)
        ax3 = plt.subplot(gs[2, 0], sharex=ax1)
    elif sharey:
        ax2 = plt.subplot(gs[1, 0], sharey=ax1)
        ax3 = plt.subplot(gs[2, 0], sharey=ax1)
    elif partial_share:
        ax2 = plt.subplot(gs[1, 0], sharex=ax1)
        ax3 = plt.subplot(gs[2, 0], sharex=ax1, sharey=ax1)
    else:
        ax2 = plt.subplot(gs[1, 0])
        ax3 = plt.subplot(gs[2, 0])

    cbax1 = plt.subplot(gs[:, 1])

    fig.subplots_adjust(wspace=0.04, hspace=0.25)

    return fig, gs, cbax1, ax1, ax2, ax3


from __future__ import print_function, division
import numpy as np
from scipy.interpolate import interp1d
from SLOPpy.subroutines.constants import *


def shift_wavelength(wave, step, rv_shift):
    wave_shift = rv_shift / speed_of_light_km + 1.00000
    return wave * wave_shift, step * wave_shift


def shift_wavelength_array(wave, rv_shift):
    wave_shift = rv_shift / speed_of_light_km + 1.00000
    return wave * wave_shift


def shift_wavelength_to_rest(wave, step, rv_shift):
    inverse_wave_shift = (-rv_shift) / speed_of_light_km + 1.00000
    return wave / inverse_wave_shift, step / inverse_wave_shift


def rebin_exact_flux(wave_in, step_in, flux_in, wave_out, step_out,
                     quadrature=False,
                     preserve_flux=True):
    """
    Previously named rebin_order
    :param wave_in:
    :param step_in:
    :param flux_in:
    :param wave_out:
    :param step_out:
    :param quadrature:
    :param preserve_flux:
    :return:

    Spectral rebinning with flux conservation
    """
    if quadrature:
        flux_in = flux_in**2.

    flux_out = np.zeros(np.shape(wave_out), dtype=np.double)
    n1 = np.size(wave_in)
    n2 = np.size(wave_out)
    ns_prv = 0

    for i in range(0, n2):
        # print i, ' of ', n2
        # Starting and ending point of the bin
        wlb = wave_out[i] - step_out[i] / 2.000
        wle = wave_out[i] + step_out[i] / 2.000

        # Normalized flux value within the bin
        fl_nm = 0.00

        # b->blue and r->red side of the original spectrum which include the bin
        # ib and ie are initialized with values close to the ones of the last iteration to save time
        ib = ns_prv
        ir = ns_prv

        for ns in range(ns_prv, n1 - 1):
            # simple algorithm to search the closest indexes near the bin boundaries
            if wave_in[ib] + step_in[ib] / 2.00 < wlb: ib += 1
            if wave_in[ir] + step_in[ir] / 2.00 < wle: ir += 1

            # when we are close to the boundary of the spectra, we stop
            if ir < ns - 3: break

        # Fail-safe checks
        if ib > ns_prv: ns_prv = ib - 3
        if ib < 0 or ir > n1: continue
        if ib > ir: continue
        if ns_prv < 0: ns_prv = 0

        # Now the true rebinning section
        if ib == ir:
            pix_s = (wle - wlb) / step_in[ib]  # fraction
            pix_e = 0.
            flux_out[i] += pix_s * flux_in[ib]
            fl_nm += pix_s
        elif ib + 1 == ir:
            pix_s = (wave_in[ib] + step_in[ib] * 0.5 - wlb) / step_in[ib]
            pix_e = (wle - (wave_in[ir] - step_in[ir] * 0.5)) / step_in[ir]
            flux_out[i] += (pix_s * flux_in[ib] + pix_e * flux_in[ir])
            fl_nm += (pix_s + pix_e)
        else:
            pix_s = (wave_in[ib] + step_in[ib] * 0.5 - wlb) / step_in[ib]
            pix_e = (wle - (wave_in[ir] - step_in[ir] * 0.5)) / step_in[ir]
            flux_out[i] += (pix_s * flux_in[ib] + pix_e * flux_in[ir])
            fl_nm += (pix_s + pix_e)
            for j in range(ib + 1, ir):
                flux_out[i] += flux_in[j]
                fl_nm += 1.00
        if (not preserve_flux) and fl_nm > 0.0:
            if quadrature:
                fl_nm *= fl_nm
            flux_out[i] /= fl_nm

    if quadrature:
        return np.sqrt(flux_out)
    else:
        return flux_out

def rebin_with_interpolation(wave_in, step_in, flux_in, wave_out, step_out,
                             quadrature=False,
                             preserve_flux=True,
                             interp_kind='cubic'):

    ndata = len(wave_in)

    normalization_factor = 1.0
    if preserve_flux:
        step_in_internal = np.ones(ndata)
        step_out_internal = np.ones(len(step_out))
    else:
        step_in_internal = step_in
        step_out_internal = step_out
        if quadrature:
            normalization_factor = (np.median(step_out) / np.median(step_in))
    if quadrature:
        flux_in = np.power(flux_in, 2.)

    wave_in_cumul = np.zeros(ndata+1)
    flux_in_cumul = np.zeros(ndata+1)

    flux_in_cumul[0] = 0.0

    wave_in_cumul[0] = wave_in[0] - step_in[0] / 2.0

    for i in range(1, ndata):
        flux_in_cumul[i] = flux_in_cumul[i - 1] + flux_in[i - 1] * step_in_internal[i - 1]
        # wave_in_cumul[i] = wave_in[i]-step_in[i]/2.
        wave_in_cumul[i] = wave_in[i] - (wave_in[i] - wave_in[i - 1]) / 2.

    flux_in_cumul[ndata] = flux_in_cumul[ndata - 1] + flux_in[ndata - 1] * step_in_internal[ndata - 1]
    wave_in_cumul[ndata] = wave_in[ndata - 1] + step_in[ndata - 1] / 2.

    flux_cumul_interp1d = interp1d(wave_in_cumul, flux_in_cumul, kind=interp_kind, bounds_error=False, fill_value=0.000)

    flux_out = (flux_cumul_interp1d(wave_out + step_out / 2.) - flux_cumul_interp1d(
        wave_out - step_out / 2.)) / step_out_internal

    if quadrature:
        return np.sqrt(flux_out) / normalization_factor
    else:
        return flux_out


def rebin_1d_to_1d(wave_in, step_in, flux_in, wave_out, step_out,
                   rv_shift=None,
                   is_error=False,
                   quadrature=False,
                   preserve_flux=True,
                   method='cubic_interpolation',
                   reference_value=None):

    if is_error:
        quadrature = True
        method='exact_flux'

    if rv_shift:
        wave_in, step_in = shift_wavelength(wave_in, step_in, rv_shift)

    if method == 'exact_flux':
        flux_out = rebin_exact_flux(wave_in, step_in, flux_in, wave_out, step_out,
                                    quadrature=quadrature, preserve_flux=preserve_flux)
    elif method == 'cubic_interpolation':
        flux_out = rebin_with_interpolation(wave_in, step_in, flux_in, wave_out, step_out,
                                            quadrature=quadrature, preserve_flux=preserve_flux, interp_kind='cubic')


    else:
        raise ValueError("method ", method, 'not supported by rebinning subroutine')


    if reference_value :
        wave_sel = (wave_out<=wave_in[0] +0.005 ) | (wave_out>=wave_in[-1] -0.005)
        flux_out[wave_sel] = reference_value

    return flux_out


def rebin_2d_to_1d(wave_in, step_in, flux_in, blaze_in, wave_out, step_out,
                   rv_shift=None,
                   is_error=False,
                   quadrature=False,
                   preserve_flux=True,
                   skip_blaze_correction=False,
                   method='cubic_interpolation',
                   reference_value=None):

    """
    :param wave_in:
    :param step_in:
    :param flux_in:
    :param blaze_in:
    :param wave_out:
    :param step_out:
    :param rv_shift:
    :param is_error:
    :param quadrature:
    :param preserve_flux:
    :param skip_blaze_correction:
    :param method:
    :return: flux_out
    """

    if rv_shift:
        wave_in, step_in = shift_wavelength(wave_in, step_in, rv_shift)

    o_axis, f_axis = np.shape(wave_in)
    n_rebin = np.size(wave_out)

    if skip_blaze_correction or blaze_in is None:
        flux_deblazed_in = flux_in
    else:
        flux_deblazed_in = flux_in / blaze_in

    if is_error:
        quadrature = True
        method = 'exact_flux'

    # Rebinning of the individual orders. We keep track of starting
    # and ending points of each order in the rebinned solution
    flux_rebin_pix = np.zeros([o_axis, n_rebin], dtype=np.double)

    counter_is = np.zeros(o_axis, dtype=np.int64) - 1
    counter_ie = np.zeros(o_axis, dtype=np.int64) - 1
    skip_order = np.ones(o_axis, dtype=bool)
    for ii in range(0, o_axis):

        counter_is[ii] = np.argmin(abs(wave_in[ii, 0] - wave_out))
        counter_ie[ii] = np.argmin(abs(wave_in[ii, -1] - wave_out))

        if wave_in[ii, 0] > np.amax(wave_out) or wave_in[ii, -1] < np.amin(wave_out):
            continue

        skip_order[ii] = False

        i = counter_is[ii]
        j = counter_ie[ii]+1

        flux_rebin_pix[ii, i:j] = rebin_1d_to_1d(wave_in[ii, :],
                                                 step_in[ii, :],
                                                 flux_deblazed_in[ii, :],
                                                 wave_out[i:j],
                                                 step_out[i:j],
                                                 quadrature=quadrature,
                                                 is_error=is_error,
                                                 preserve_flux=preserve_flux,
                                                 method=method,
                                                 reference_value=reference_value)

    flux_out = np.zeros(n_rebin, dtype=np.double)
    if reference_value:
        flux_out += reference_value

    if quadrature or is_error:
        flux_rebin_pix = np.power(flux_rebin_pix, 2)
        flux_out[counter_is[0]:counter_ie[0]] = flux_rebin_pix[0, counter_is[0]:counter_ie[0]]

        for ii in range(1, o_axis):
            if skip_order[ii]: continue

            p_ie = counter_ie[ii - 1]
            j_is = counter_is[ii]
            j_ie = counter_ie[ii] + 1 # adding one because it is used in interval definition - Python quirks

            if p_ie > j_is:
                nr_joint = float(p_ie - j_is)
                ij = np.arange(j_is, p_ie, 1, dtype=np.int64)
                ij_fraction = np.power((ij-j_is) / nr_joint, 4)
                flux_out[ij] = flux_rebin_pix[ii,ij] * ij_fraction + flux_rebin_pix[ii-1,ij] * (1. - ij_fraction)
                flux_out[p_ie:j_ie] = flux_rebin_pix[ii, p_ie:j_ie]
            else:
                flux_out[j_is:j_ie] = flux_rebin_pix[ii, j_is:j_ie]

        return np.sqrt(flux_out)

    else:
        flux_out[counter_is[0]:counter_ie[0]] = flux_rebin_pix[0, counter_is[0]:counter_ie[0]]
        for ii in range(1, o_axis):
            if skip_order[ii]: continue

            p_ie = counter_ie[ii - 1]
            j_is = counter_is[ii]
            j_ie = counter_ie[ii] + 1

            if p_ie > j_is:
                nr_joint = float(p_ie - j_is)
                ij = np.arange(j_is, p_ie, 1, dtype=np.int64)
                ij_fraction = np.power((ij-j_is) / nr_joint, 2.)
                flux_out[ij] = flux_rebin_pix[ii,ij] * ij_fraction + flux_rebin_pix[ii-1,ij] * (1. - ij_fraction)
                flux_out[p_ie:j_ie] = flux_rebin_pix[ii, p_ie:j_ie]
            else:
                flux_out[j_is:j_ie] = flux_rebin_pix[ii, j_is:j_ie]
        return flux_out


def rebin_1d_to_2d(wave_in, step_in, flux_in, wave_out, step_out,
                   rv_shift=None,
                   is_error=False,
                   quadrature=False,
                   preserve_flux=True,
                   method='cubic_interpolation',
                   reference_value=None):
    """

    :param wave_in:
    :param step_in:
    :param flux_in:
    :param wave_out:
    :param step_out:
    :param rv_shift:
    :param is_error:
    :param quadrature:
    :param preserve_flux:
    :param method:
    :return:
    """

    if rv_shift:
        wave_in, step_in = shift_wavelength(wave_in, step_in, rv_shift)

    if is_error:
        quadrature = True
        method='exact_flux'

    o_axis, f_axis = np.shape(wave_out)
    flux_out = np.zeros([o_axis, f_axis], dtype=np.double)
    if reference_value:
        flux_out += reference_value

    for ii in range(0, o_axis):
        flux_out[ii, :]= rebin_1d_to_1d(wave_in,
                                        step_in,
                                        flux_in,
                                        wave_out[ii, :],
                                        step_out[ii, :],
                                        quadrature=quadrature,
                                        is_error=is_error,
                                        preserve_flux=preserve_flux,
                                        method=method,
                                        reference_value=reference_value)


    return flux_out


def rebin_2d_to_2d(wave_in, step_in, flux_in, wave_out, step_out,
                   rv_shift=None,
                   is_error=False,
                   quadrature=False,
                   preserve_flux=True,
                   method='cubic_interpolation',
                   reference_value=None):

    """
    :param wave_in: 1D wavelength array of the input spectrum
    :param step_in: 1D step-size array of the input spectrum
    :param flux_in: 1D flux array of the input spectrum
    :param wave_out: 2D (order-by-order) wavelength array of the rebinned spectrum
    :param step_out: 2D (order-by-order) step-size array of the rebinned spectrum
    :param rv_shift:
    :param is_error:
    :param quadrature:
    :param preserve_flux:
    :param method:
    :return: flux_out, 2D (order-by-order) flux array, with same size as wave_out
    """

    if rv_shift:
        wave_in, step_in = shift_wavelength(wave_in, step_in, rv_shift)

    if is_error:
        quadrature = True
        method='exact_flux'

    o_axis_input, f_axis_input = np.shape(wave_in)
    o_axis, f_axis = np.shape(wave_out)

    if o_axis_input != o_axis:
        raise ValueError("Mismatch between input and output number of orders in rebin_2d_to_2d")

    flux_out = np.zeros([o_axis, f_axis], dtype=np.double)
    if reference_value:
        flux_out += reference_value

    for ii in range(0, o_axis):
        flux_out[ii, :] = rebin_1d_to_1d(wave_in[ii, :],
                                         step_in[ii, :],
                                         flux_in[ii, :],
                                         wave_out[ii, :],
                                         step_out[ii, :],
                                         quadrature=quadrature,
                                         is_error=is_error,
                                         preserve_flux=preserve_flux,
                                         method=method,
                                         reference_value=reference_value)

    return flux_out


from SLOPpy.subroutines.common import *
import os

from SLOPpy.subroutines.mcmc_fit_functions import *
from SLOPpy.subroutines.math_functions import interpolate2d_grid_nocheck

#from SLOPpy.subroutines.interpol import interpolate1d_grid_nocheck

from multiprocessing import Pool
import emcee
import time

# define theta pams
def define_theta_array(model_case,line_iter_dict, planet_dict, n_jitter, allow_emission=False):

    pams_dict = {}  # dictionary containing the index of a given parameter
    pams_list = []  # list wirth the parameter names ordered according to their index
    boundaries = np.empty([0, 2])  # boundaries for MCMC / nested sampling
    theta_start = np.empty(0)  # starting point for MCMC
    lines_center = np.empty(0)  # laboratory wavelength of spectral lines
    pam_index = 0  # keep track of the number of variables

    for line_key, line_val in line_iter_dict['lines'].items():
        pam_name = line_key + '_contrast'
        pams_dict[pam_name] = pam_index
        pams_list.append(pam_name)
        if allow_emission:
            boundaries = np.append(boundaries, [[-0.20, 0.20]], axis=0)
        else:
            boundaries = np.append(boundaries, [[0.00, 0.20]], axis=0)
        theta_start = np.append(theta_start, 0.010)
        pam_index += 1

        lines_center = np.append(lines_center, line_val)

        """ skip the inclusion of FWHM as a free parameter for each line
            if the shared FWHM is selected
        """

        if model_case in [0, 1, 2, 3, 10, 11, 14, 20, 21, 24]:
            # if not line_iter_dict['fit_parameters']['shared_fwhm']:
            pam_name = line_key + '_fwhm (km/s)'
            pams_dict[pam_name] = pam_index
            pams_list.append(pam_name)
            boundaries = np.append(boundaries, [[0.00, 100.00]], axis=0)
            theta_start = np.append(theta_start, 5.0)
            pam_index += 1

        # if line_iter_dict['fit_parameters']['fixed_separation']: continue
        # if not line_iter_dict['fit_parameters']['lines_shift']: continue

        if model_case in [0, 2, 10, 12, 20, 22]:
            pam_name = line_key + '_winds  (km/s)'
            pams_dict[pam_name] = pam_index
            pams_list.append(pam_name)
            boundaries = np.append(boundaries, [[-25.00, 25.00]], axis=0)
            theta_start = np.append(theta_start, 0.00)
            pam_index += 1

    if model_case in [12, 13, 15, 22, 23, 25]:
        # if line_iter_dict['fit_parameters']['shared_fwhm']:
        pam_name = 'shared_fwhm  (km/s)'
        pams_dict[pam_name] = pam_index
        pams_list.append(pam_name)
        boundaries = np.append(boundaries, [[0.000, 100.00]], axis=0)
        theta_start = np.append(theta_start, 5.000)
        pam_index += 1

    if model_case in [11, 13, 21, 23]:
        # if line_iter_dict['fit_parameters']['fixed_separation'] and line_iter_dict['fit_parameters']['lines_shift']:
        pam_name = 'shared_winds  (km/s)'
        pams_dict[pam_name] = pam_index
        pams_list.append(pam_name)
        boundaries = np.append(boundaries, [[-25.0, 25.0]], axis=0)
        theta_start = np.append(theta_start, 0.000)
        pam_index += 1

    if model_case in [0, 1, 10, 11, 12, 13, 14, 15]:
        pams_dict['rp_factor'] = pam_index
        pams_list.append('rp_factor')
        boundaries = np.append(boundaries, [[0.5, 2.0]], axis=0)
        theta_start = np.append(theta_start, 1.0)
        pam_index += 1

    pams_dict['K_planet  (km/s)'] = pam_index
    pams_list.append('K_planet  (km/s)')
    #boundaries = np.append(boundaries,
    #                        [[-300., planet_dict['RV_semiamplitude']
    #                            [0] + 300.]],
    #                        axis=0)
    '''
    boundaries = np.append(boundaries,
                            [[planet_dict['RV_semiamplitude'][0] - 75.,
                              planet_dict['RV_semiamplitude'][0] + 75.]],
                            axis=0)

    '''
    boundaries = np.append(boundaries,
                            [[0.,
                              300.]],
                            axis=0)
    
    theta_start = np.append(
        theta_start, planet_dict['RV_semiamplitude'][0] / 1000.0)
    pam_index += 1


    for i_j in range(0,n_jitter):
        pam_name = 'jitter_' + repr(i_j)
        pams_dict[pam_name] = pam_index
        pams_list.append(pam_name)
        boundaries = np.append(boundaries, [[10**(-12), 0.05]], axis=0)
        theta_start = np.append(theta_start, 10**(-11))
        pam_index += 1

    return lines_center, pams_dict, pams_list, boundaries, theta_start




def emcee_lines_fit_functions(model_case,
                              wave_meshgrid,
                              transmission_spec,
                              transmission_spec_err,
                              clv_rm_radius,
                              clv_rm_grid,
                              planet_RVsinusoid,
                              lines_center,
                              jitter_index,
                              priors_dict,
                              theta_start, boundaries, ndim, nwalkers, ngen, nsteps, nthin, ncpus):

    os.environ["OMP_NUM_THREADS"] = "1"

    #""" Avoid starting values out of boundaries """
    #for nd in range(0, ndim):
    #    sel = (point_star[:,nd] <= boundaries[nd,0]) | (point_star[:,nd] >= boundaries[nd,1])
    #    point_star[sel,nd] = theta_start[nd]


    """
    print(np.shape(wave_append))
    print(np.shape(flux_append))
    print(np.shape(ferr_append))
    print(np.shape(nobs_append))
    print(np.shape(clv_rm_radius))
    print(np.shape(clv_rm_grid))
    print(np.shape(rvsys_PRF2ORF_append))
    print(np.shape(planet_RVsinusoid_append))
    print(np.shape(lines_center))
    """

    model_dictionaries = {
         0: logprob_case00,
         1: logprob_case01,
         2: logprob_case02,
         3: logprob_case03,
        10: logprob_case10,
        11: logprob_case11,
        12: logprob_case12,
        13: logprob_case13,
        14: logprob_case14,
        15: logprob_case15,
        20: logprob_case20,
        21: logprob_case21,
        22: logprob_case22,
        23: logprob_case23,
        24: logprob_case24,
        25: logprob_case25,
    }

    logprob = model_dictionaries[model_case]

    try:
        #from pyde.de import DiffEvol
        from pytransit.utils.de import DiffEvol
        use_pyde = True
    except ImportError:
        print('   Warnign: PyDE is not installed, random initialization point')
        use_pyde = False

    if ngen <= 1 : use_pyde = False

    """ R_p is fixed to 1.0 """
    if model_case in [2, 3, 20, 21, 22, 23, 24, 25]:
        clv_model = interpolate2d_grid_nocheck(1.000, clv_rm_radius, clv_rm_grid)
        args_input = (boundaries,
                        wave_meshgrid,
                        transmission_spec,
                        transmission_spec_err,
                        clv_model,
                        planet_RVsinusoid,
                        lines_center,
                        jitter_index,
                        priors_dict)
    else:
        args_input = (boundaries,
                        wave_meshgrid,
                        transmission_spec,
                        transmission_spec_err,
                        clv_rm_radius,
                        clv_rm_grid,
                        planet_RVsinusoid,
                        lines_center,
                        jitter_index,
                        priors_dict)


    if use_pyde:
        start = time.time()

        with Pool() as pool:

            de = DiffEvol(
                    logprob,
                    boundaries,
                    nwalkers,
                    maximize=True,
                    pool=pool,
                    args=args_input)

            de.optimize(ngen)

            end = time.time()
            print("PyDE global optimization took {0:.1f} seconds".format( end - start))

            theta_start = np.median(de.population, axis=0)
            point_start = de.population
    else:
        point_start = theta_start  + 1e-4 * np.abs(np.random.randn(nwalkers, ndim))
        point_start[0, :] = theta_start

    start = time.time()
    with Pool(processes=ncpus) as pool:

        sampler = emcee.EnsembleSampler(nwalkers,
                                        ndim,
                                        logprob,
                                        args=args_input,
                                        pool=pool)

        population, prob, state = sampler.run_mcmc(point_start,
                                                   nsteps,
                                                   thin=nthin,
                                                   progress=True)

    end = time.time()
    print()
    print("emcee MCMC optimization took {0:.1f} seconds".format(end - start))

    return population, sampler.chain, sampler.lnprobability, point_start


def return_model(model_case,
                theta,
                wave_meshgrid,
                clv_rm_radius,
                clv_rm_grid,
                planet_RVsinusoid,
                lines_center,
                jitter_index):

    ndim = len(theta)
    boundaries = np.empty([ndim, 2])
    boundaries[:,0] = theta - 1.
    boundaries[:,1] = theta + 1.
    transmission_spec = np.ones(np.shape(wave_meshgrid))
    transmission_spec_err = np.ones(np.shape(wave_meshgrid))

    model_dictionaries = {
         0: logprob_case00,
         1: logprob_case01,
         2: logprob_case02,
         3: logprob_case03,
        10: logprob_case10,
        11: logprob_case11,
        12: logprob_case12,
        13: logprob_case13,
        14: logprob_case14,
        15: logprob_case15,
        20: logprob_case20,
        21: logprob_case21,
        22: logprob_case22,
        23: logprob_case23,
        24: logprob_case24,
        25: logprob_case25,
    }

    logprob = model_dictionaries[model_case]

    if model_case in [2, 3, 20, 21, 22, 23, 24, 25]:
        clv_model = interpolate2d_grid_nocheck(1.000, clv_rm_radius, clv_rm_grid)
        lines_model, _, lines_array, planet_K, planet_R, jitter = logprob(
                theta,
                boundaries,
                wave_meshgrid,
                transmission_spec,
                transmission_spec_err,
                clv_model,
                planet_RVsinusoid,
                lines_center,
                jitter_index,
                {},
                return_models=True)

    else:
        lines_model, clv_model, lines_array, planet_K, planet_R, jitter = logprob(
                theta,
                boundaries,
                wave_meshgrid,
                transmission_spec,
                transmission_spec_err,
                clv_rm_radius,
                clv_rm_grid,
                planet_RVsinusoid,
                lines_center,
                jitter_index,
                {},
                return_models=True)


    return lines_model, clv_model, lines_array, planet_K, planet_R, jitter


def emcee_flatten_median(population, sampler_chain, sampler_lnprobability,  nburnin, nthin, nwalkers):

    flat_chain = emcee_flatchain(sampler_chain, nburnin, nthin)
    flat_lnprob, _ = emcee_flatlnprob(
            sampler_lnprobability, nburnin, nthin, population, nwalkers)

    lnprob_med = compute_value_sigma(flat_lnprob)
    chain_med = compute_value_sigma(flat_chain)
    chain_MAP, lnprob_MAP = pick_MAP_parameters(flat_chain, flat_lnprob)

    #n_samplings, n_pams = np.shape(flat_chain)
    return flat_chain, flat_lnprob, chain_med, chain_MAP, lnprob_med, lnprob_MAP

def emcee_compute_BIC_AIC(lnprob_med, lnprob_MAP, ndata, ndim):

    print()
    print(' LN posterior: {0:12f}   {1:12f} {2:12f} (15-84 p)  MAP:  {0:12f}'.format(
        lnprob_med[0], lnprob_med[2], lnprob_med[1], lnprob_MAP))

    BIC = -2.0 * lnprob_med[0] + np.log(ndata) * ndim
    AIC = -2.0 * lnprob_med[0] + 2.0 * ndim
    AICc = AIC + (2.0 + 2.0 * ndim) * ndim / (ndata - ndim - 1.0)

    print()
    print(' Median BIC    = {}'.format(BIC))
    print(' Median AIC    = {}'.format(AIC))
    print(' Median AICc   = {}'.format(AICc))

    BIC_map = -2.0 * lnprob_MAP + np.log(ndata) * ndim
    AIC_map = -2.0 * lnprob_MAP + 2.0 * ndim
    AICc_map = AIC + (2.0 + 2.0 * ndim) * ndim / (ndata - ndim - 1.0)

    print()
    print(' MAP BIC       = {}'.format(BIC_map))
    print(' MAP AIC       = {}'.format(AIC_map))
    print(' MAP AICc      = {}'.format(AICc_map))


def emcee_burnin_check(chain, nburnin, nthin, nwalkers=False):
    nburn = int(nburnin / nthin)
    modified = False

    if not nwalkers:
        _, d, _ = np.shape(chain)
    else:
        v1, v2 = np.shape(chain)
        if v1 == nwalkers:
            d = v2
        else:
            d = v1

    if nburn >= d * 0.9:
        nburn = int(d / 4)
        modified = True

    return nburn, modified

def emcee_flatchain(chain, nburnin, nthin):
    """flattening of the emcee chains with removal of burn-in"""
    nburn, _ = emcee_burnin_check(chain, nburnin, nthin)
    s = chain[:, nburn:, :].shape
    return chain[:, nburn:, :].reshape(s[0] * s[1], s[2])


def emcee_flatlnprob(lnprob, nburnin, nthin, population, nwalkers):

    nburn, _  = emcee_burnin_check(lnprob, nburnin, nthin, nwalkers)

    v1, v2 = np.shape(lnprob)
    if v1 == nwalkers:
        s = lnprob[:, nburn:].shape
        return lnprob[:, nburn:].reshape(s[0] * s[1]), lnprob.T
    else:
        s = lnprob[nburn:, :].shape
        return lnprob[nburn:, :].reshape(s[0] * s[1]), lnprob

def GelmanRubin_v2(sampler_chain):
    """
    :param chain_T:
    :return:
    """

    """
    from http://joergdietrich.github.io/emcee-convergence.html
    """
    ssq = np.var(sampler_chain, axis=1, ddof=1)
    W = np.mean(ssq, axis=0)
    theta_b = np.mean(sampler_chain, axis=1)
    theta_bb = np.mean(theta_b, axis=0)
    m = sampler_chain.shape[0] * 1.0
    n = sampler_chain.shape[1] * 1.0
    B = n / (m - 1) * np.sum((theta_bb - theta_b) ** 2, axis=0)
    var_theta = (n - 1) / n * W + 1 / n * B
    Rhat = np.sqrt(var_theta / W)
    return Rhat


def compute_value_sigma(samples):
    if np.size(np.shape(samples)) == 1:
        sample_med = np.zeros(3)
        #sample_tmp = np.percentile(samples, [15.865, 50, 84.135], axis=0)
        sample_tmp = np.percentile(samples[np.isfinite(samples)], [15.865, 50, 84.135], axis=0)

        sample_med[0] = sample_tmp[1]
        sample_med[1] = sample_tmp[2] - sample_tmp[1]
        sample_med[2] = sample_tmp[1] - sample_tmp[0]

    elif np.size(np.shape(samples)) == 2:
        #sample_med = np.asarray(list(map(lambda v: (v[1], v[2] - v[1], v[1] - v[0]),
                                   #zip(*np.percentile(samples, [15.865, 50, 84.135], axis=0)))))
        sample_med = np.zeros((samples.shape[1],3))
        for k in range(samples.shape[1]):
            ttt = samples[:,k]
            sample_tmp = np.percentile(ttt[np.isfinite(ttt)], [15.865, 50, 84.135], axis=0)
            sample_med[k,0] = sample_tmp[1]
            sample_med[k,1] = sample_tmp[2] - sample_tmp[1]
            sample_med[k,2] = sample_tmp[1] - sample_tmp[0]


    else:
        print('ERROR!!! ')
        return None
    return sample_med

def pick_MAP_parameters(samples, lnprob):

    indmax = np.argmax(lnprob)
    if np.size(np.shape(samples)) == 1:
        return samples[indmax], lnprob[indmax]
    elif np.size(np.shape(samples)) == 2:
        return samples[indmax, :], lnprob[indmax]
    else:
        print('ERROR!!! ')
        return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import division # no more "zero" integer division bugs!:P
import numpy as np # array


# radiants, degrees conversions etc.
pi = 4.*np.arctan(1.)
dpi = 2.*pi
deg2rad = pi/180.
rad2deg = 180./pi

# various
TOLERANCE = np.finfo(np.float64(1.0)).eps
d2s = 86400. # seconds in a day  =  24h  =  86400 s
d2m = 1440.  # min in a day  =  1440. min

# masses conversions
Msmer = 6.0236e6 # Msun to Mmer
Mmers = 1./Msmer # Mmer to Msun
Msven = 4.08523719e5 # Msun to Mven
Mvens = 1./Msven   #  Mven to Msun
Msear = 332946.0487 # Msun to Mear
Mears = 1./Msear  #  Mear to Msun
Msmar = 3.09870359e6 # Msun to Mmar
Mmars = 1./Msmar   #  Mmar to Msun
Msjup = 1.047348644e3 # Msun to Mjup
Mjups = 1./Msjup    #  Mjup to Msun
Mssat = 3.4979018e3 # Msun to Msat
Msats = 1./Mssat  #  Msat to Msun
Msura = 2.290298e4 # Msun to Mura
Muras = 1./Msura #  Mura to Msun
Msnep = 1.941226e4 # Msun to Mnep
Mneps = 1./Msnep #  Mnep to Msun

Mejup = Mears * Msjup # Mear to Mjup
Mjear = Mjups * Msear # Mjup to Mear

# masses of Solar System objects
Msun = 1.9884e30 # Sun mass in kg
Mmer = Msun*Mmers #  Mercury mass in kg
Mven = Msun*Mvens #  Venus mass in kg
Mear = 5.9722e24 # Earth mass in kg
Mmar = Msun*Mmars #  Mars mass in kg
Mjup = Msun*Mjups #  Jupiter mass in kg
Msat = Msun*Msats #  Saturn mass in kg
Mura = Msun*Muras #  Uranus mass in kg
Mnep = Msun*Mneps #  Neptune mass in kg

# radii of Solar System objects
Rsun = 696000. #  Sun radius in km
Rmer = 2439.7  #  Mercury radius in km
Rven = 6051.8  #  Venus radius in km
Rear = 6378.1366 # Earth radius in km
Rmar = 3396.19 #  Mars radius in km
Rjup = 71492.  #  Jupiter radius in km
Rsat = 60268.  #  Saturn radius in km
Rura = 25559.  #  Uranus radius in km
Rnep = 24764.  #  Neptune radius in km
Rplu = 1195.   #  Pluto radius in km
#
Rsjup = Rsun/Rjup # Rsun to Rjup
Rjups = Rjup/Rsun # Rjup to Rsun
Rsear = Rsun/Rear # Rsun to Rear
Rears = Rear/Rsun # Rear to Rsun
Rsnep = Rsun/Rnep # Rsun to Rnep
Rneps = Rnep/Rsun # Rnep to Rsun
#
Rejup = Rear/Rjup # Rearth to Rjupiter
Rjear = Rjup/Rear # Rjupiter to Rearth


# astronomical constants
AU = 149597870700. #Astronomical Unit in meters
kappa = 0.01720209895 # Gaussian gravitational constant
Giau = kappa*kappa # G [AU^3/Msun/d^2]
Gsi = 6.67428e-11 #Gravitational Constants in SI system [m^3/kg/s^2]
Gaumjd = Gsi*d2s*d2s*Mjup/(AU**3) # G in [AU,Mjup,day]
speed = 299792458. # speed of light (c) in [m/s]
speedaud = speed*d2s/AU # speed of light in [AU/d]
pc2AU = 206264.806 


# others
RsunAU = (Rsun*1.e3)/AU #Sun radius in AU
RjupAU = (Rjup*1.e3)/AU #Jupiter radius in AU

MJD = 2400000.5 # MJD ref time to convert to JD

speed_of_light = speed

sigma2fwhm = 2. * np.sqrt(2. * np.log(2))
speed_of_light_km = speed / 1000. 


from __future__ import print_function, division
from SLOPpy.subroutines.common import *
from SLOPpy.subroutines.constants import *
from SLOPpy.subroutines.spectral_subroutines import *
from SLOPpy.subroutines.io_subroutines import *
from SLOPpy.subroutines.fit_subroutines import *
from SLOPpy.subroutines.shortcuts import *
from SLOPpy.subroutines.math_functions import *
from SLOPpy.subroutines.plot_subroutines import *
from SLOPpy.subroutines.bayesian_emcee import *
# from SLOPpy.subroutines.rebin_subroutines import *
from scipy.signal import savgol_filter

__all__ = ['compute_transmission_binned_mcmc','compute_transmission_binned_mcmc_iterative',
           'plot_transmission_binned_mcmc','plot_transmission_binned_mcmc_iterative']

subroutine_name = 'transmission_binned_mcmc'


def compute_transmission_binned_mcmc_iterative(config_in, lines_label):

    pca_parameters = from_config_get_pca_parameters(config_in)
    for it in range(0, pca_parameters.get('iterations',5)):
        compute_transmission_binned_mcmc(config_in, lines_label, reference='planetRF', pca_iteration=it)



def plot_transmission_binned_mcmc_iterative(config_in, lines_label, night_input='', reference='planetRF'):

    pca_parameters = from_config_get_pca_parameters(config_in)
    for it in range(0, pca_parameters.get('iterations',5)):
        plot_transmission_binned_mcmc(config_in, lines_label, night_input=night_input, reference=reference, pca_iteration=it)





def compute_transmission_binned_mcmc(config_in, lines_label, reference='planetRF', pca_iteration=-1):

    night_dict = from_config_get_nights(config_in)
    planet_dict = from_config_get_planet(config_in)
    star_dict = from_config_get_star(config_in)
    clv_rm_dict = from_config_get_clv_rm(config_in)

    shared_data = load_from_cpickle('shared', config_in['output'])

    """ selection of those parameters that are specific of the spectral line(s)
        under analysis
    """

    spectral_lines = from_config_get_spectral_lines(config_in)
    lines_dict = spectral_lines[lines_label]

    sampler_pams = lines_dict['sampler_parameters']
    sampler_name = sampler_pams.get('sampler_name', 'emcee')

    # TODO reference as input parameter
    reference = 'planetRF'

    """
    - case  0: only one spectral line, default line parameters are contrast, FWHM, rv_shift
    - case  1: only one spectral line, no winds
    - case  2: only one spectral line, no planetary radius dependance
    - case  3: only one spectral line, no winds and no planetary radius dependance
    - case 10: more than one spectral lines, all line parameters are free and independent
    - case 11: more than one spectral lines, all lines are affected by the same wind
    - case 12: more than one spectral lines, all lines have same FWHM
    - case 13: more than one spectral lines, all lines are affected by the same wind and have same FWHM
    - case 14: more than one spectral lines, no winds
    - case 15: more than one spectral lines, no winds, all lines have same FWHM
    - case 20: more than one spectral lines, no Rp dependance, all line parameters are free and independent
    - case 21: more than one spectral lines, no Rp dependance, all lines are affected by the same wind
    - case 22: more than one spectral lines, no Rp dependance, all lines have same FWHM
    - case 23: more than one spectral lines, no Rp dependance, all lines are affected by the same wind and have same FWHM
    - case 24: more than one spectral lines, no Rp dependance, no winds
    - case 25: more than one spectral lines, no Rp dependance, no winds, all lines have same FWHM

                free_Rp     free_winds  shared_winds    shared_FWHM
    - case  0:  True        True        False           False               DEFAULT for single line
    - case  1:  True        False       False           False
    - case  2:  False       True        False           False
    - case  3:  False       False       False           False


    - case 10:  True        True        False           False               DEFAULT for multiple lines
    - case 11:  True        True        True            False
    - case 12:  True        True        False           True
    - case 13:  True        True        True            True
    - case 14:  True        False       False           False
    - case 15:  True        False       False           True

    - case 20:  False       True        False           False
    - case 21:  False       True        True            False
    - case 22:  False       True        False           True
    - case 23:  False       True        True            True
    - case 24:  False       False       False           False
    - case 25:  False       False       False           True
    """

    model_case = 10

    norm_dict = lines_dict.get('normalization', clv_rm_dict.get('normalization', {}))
    norm_pams={}
    norm_pams['normalize_transmission'] = norm_dict.get('normalize_transmission', True)
    norm_pams['normalization_model'] = norm_dict.get('normalization_model', 'polynomial')

    """ Normalization parameters for polynomial model"""
    norm_pams['model_poly_degree'] = norm_dict.get('model_poly_degree', 2)
    norm_pams['spectra_poly_degree'] = norm_dict.get('spectra_poly_degree', 2)
    norm_pams['lower_threshold'] = norm_dict.get('lower_threshold', 0.950)
    norm_pams['percentile_selection'] = norm_dict.get('percentile_selection', 10)

    """ Normalization parameters using Savitzky-Golay filter"""
    norm_pams['window_length'] = norm_dict.get('window_length', 101)
    norm_pams['polyorder'] = norm_dict.get('polyorder', 3)
    norm_pams['mode'] = norm_dict.get('mode', 'nearest')
    norm_pams['cval'] = norm_dict.get('cval', 1.0)
    norm_pams['normalize_rebinned'] = norm_dict.get('normalize_rebinned', False)

    # Added back-compatibility to old or "wrong" keys
    clv_rm_correction = lines_dict.get('clv_rm_correction', True)

    fit_pams = lines_dict['fit_parameters']
    free_Rp = fit_pams.get('free_Rp', True) \
        and fit_pams.get('free_planet_radius', True) \
        and clv_rm_correction
    free_winds = fit_pams.get('free_winds', True) \
        and fit_pams.get('free_offset', True)
    shared_winds = fit_pams.get('shared_winds', False) \
        or fit_pams.get('shared_offset', False)
    shared_FWHM = fit_pams.get('shared_FWHM', False) \
        or fit_pams.get('shared_fwhm', False)

    prior_dict = fit_pams.get('priors', {}) \
        or fit_pams.get('priors', {})

    allow_emission = fit_pams.get('allow_emission', False)



    if len(lines_dict['lines']) < 2:
        if free_Rp is True and free_winds is True:
            model_case = 0
        if free_Rp is True and free_winds is False:
            model_case = 1
        if free_Rp is False and free_winds is True:
            model_case = 2
        if free_Rp is False and free_winds is False:
            model_case = 3
    else:
        if free_Rp is True:
            if free_winds is True:
                if shared_winds is False and shared_FWHM is False:
                    model_case = 10
                if shared_winds is True and shared_FWHM is False:
                    model_case = 11
                if shared_winds is False and shared_FWHM is True:
                    model_case = 12
                if shared_winds is True and shared_FWHM is True:
                    model_case = 13
            else:
                if shared_winds is False and shared_FWHM is False:
                    model_case = 14
                if shared_winds is False and shared_FWHM is True:
                    model_case = 15
        else:
            if free_winds is True:
                if shared_winds is False and shared_FWHM is False:
                    model_case = 20
                if shared_winds is True and shared_FWHM is False:
                    model_case = 21
                if shared_winds is False and shared_FWHM is True:
                    model_case = 22
                if shared_winds is True and shared_FWHM is True:
                    model_case = 23
            else:
                if shared_winds is False and shared_FWHM is False:
                    model_case = 24
                if shared_winds is False and shared_FWHM is True:
                    model_case = 25

    jitter_flag = fit_pams.get('jitter', True)
    pyde_flag = fit_pams.get('pyde', True)

    print()
    print('      free_Rp:         (default: True)  ', free_Rp)
    print('      free_winds:      (default: True) ', free_winds)
    print('      shared_winds:    (default: False) ', shared_winds)
    print('      shared_FWHM:     (default: False) ', shared_FWHM)
    print('      jitter:          (default: True)  ', jitter_flag)
    print('      # lines:        ', len(lines_dict['lines']))
    print('      model_case:     ', model_case)

    """ parameters list:
        to be updated


    pams_dict = {}  # dictionary containing the index of a given parameter
    pams_list = []  # list with the parameter names ordered according to their index
    boundaries = np.empty([0, 2])  # boundaries for MCMC / nested sampling
    theta_start = np.empty(0)  # starting point for MCMC
    lines_center = np.empty(0)  # laboratory wavelength of spectral lines
    pam_index = 0  # keep track of the number of variables

    for line_key, line_val in lines_dict['lines'].items():
        pam_name = line_key + '_contrast'
        pams_dict[pam_name] = pam_index
        pams_list.append(pam_name)
        boundaries = np.append(boundaries, [[0.00, 1.00]], axis=0)
        theta_start = np.append(theta_start, 0.010)
        pam_index += 1

        lines_center = np.append(lines_center, line_val)

        # skip the inclusion of FWHM as a free parameter for each line
            if the shared FWHM is selected
        #

        if model_case in [0, 1, 2, 3, 10, 11, 14, 20, 21, 24]:
            # if not lines_dict['fit_parameters']['shared_fwhm']:
            pam_name = line_key + '_fwhm'
            pams_dict[pam_name] = pam_index
            pams_list.append(pam_name)
            boundaries = np.append(boundaries, [[0.00, 150.00]], axis=0)
            theta_start = np.append(theta_start, 5.0)
            pam_index += 1

        # if lines_dict['fit_parameters']['fixed_separation']: continue
        # if not lines_dict['fit_parameters']['lines_shift']: continue

        if model_case in [0, 2, 10, 12, 20, 22]:
            pam_name = line_key + '_winds'
            pams_dict[pam_name] = pam_index
            pams_list.append(pam_name)
            boundaries = np.append(boundaries, [[-5.00, 5.00]], axis=0)
            theta_start = np.append(theta_start, 0.00)
            pam_index += 1

    if model_case in [12, 13, 15, 22, 23, 25]:
        # if lines_dict['fit_parameters']['shared_fwhm']:
        pam_name = 'shared_fwhm'
        pams_dict[pam_name] = pam_index
        pams_list.append(pam_name)
        boundaries = np.append(boundaries, [[0.000, 150.00]], axis=0)
        theta_start = np.append(theta_start, 5.000)
        pam_index += 1

    if model_case in [11, 13, 21, 23]:
        # if lines_dict['fit_parameters']['fixed_separation'] and lines_dict['fit_parameters']['lines_shift']:
        pam_name = 'shared_winds'
        pams_dict[pam_name] = pam_index
        pams_list.append(pam_name)
        boundaries = np.append(boundaries, [[-5.0, 5.0]], axis=0)
        theta_start = np.append(theta_start, 0.000)
        pam_index += 1

    if model_case in [0, 1, 10, 11, 12, 13, 14, 15]:
        pams_dict['rp_factor'] = pam_index
        pams_list.append('rp_factor')
        boundaries = np.append(boundaries, [[0.5, 2.0]], axis=0)
        theta_start = np.append(theta_start, 1.0)
        pam_index += 1

    pams_dict['K_planet'] = pam_index
    pams_list.append('K_planet')
    boundaries = np.append(boundaries,
                            [[-300., planet_dict['RV_semiamplitude']
                                [0]+ 300.]],
                            axis=0)
    theta_start = np.append(
        theta_start, planet_dict['RV_semiamplitude'][0])
    pam_index += 1

    pam_name = 'jitter'
    pams_dict[pam_name] = pam_index
    pams_list.append(pam_name)
    boundaries = np.append(boundaries, [[10**(-12), 0.01]], axis=0)
    theta_start = np.append(theta_start, 10**(-11))
    pam_index += 1

    for ii in range(0, pam_index):
        print(pams_list[ii], '    ', boundaries[ii, :],
                '    ', theta_start[ii])

    ndim = pam_index

    """

    for night in night_dict:

        print()
        print("transmission_mcmc               Night: {0:s}".format(night))



        """ Retrieving the list of observations"""
        lists = load_from_cpickle('lists', config_in['output'], night)

        observational_pams = load_from_cpickle(
            'observational_pams', config_in['output'], night)

        ### Moved here to retrieve infomration about wheter PCA correction has been performed or not
        preparation_input = load_from_cpickle('transmission_preparation', config_in['output'], night)

        if preparation_input.get('pca_output', False):
            if pca_iteration >= 0:
                it_string = str(pca_iteration).zfill(2)
            else:
                it_string = str(preparation_input.get('ref_iteration', 0)).zfill(2)
            preparation = preparation_input[it_string]
        else:
            preparation = preparation_input
            it_string = ''

        """ This need to be checked only once, so it's ok to take the output of the last night
            and propagate it to the rest of subroutine
        """
        if len(it_string) > 0:
            pca_output = True
        else:
            pca_output = False

        try:
            mcmc_data = load_from_cpickle(subroutine_name + '_data', config_in['output'], night, lines_label, it_string)

            clv_rm_radius = mcmc_data['clv_rm_radius']
            clv_rm_grid = mcmc_data['clv_rm_grid']
            transmission_spec = mcmc_data['transmission_spec']
            transmission_spec_err = mcmc_data['transmission_spec_err']
            wave_meshgrid = mcmc_data['wave_meshgrid']
            time_meshgrid = mcmc_data['time_meshgrid']
            planet_RVsinusoid = mcmc_data['planet_RVsinusoid']
            jitter_index = mcmc_data['jitter_index']
            n_jitter = mcmc_data['n_jitter']

            print("   Loading MCMC data array for lines {0:s}, night: {1:s}".format(
                lines_label, night))

        except FileNotFoundError:

            print("  Computing MCMC data array for lines {0:s}, night: {1:s}".format(
                lines_label, night))

            calib_data = load_from_cpickle(
                'calibration_fibA', config_in['output'], night)
            input_data = retrieve_observations(
                config_in['output'], night, lists['observations'])

            if clv_rm_correction:
                try:
                    clv_rm_models = load_from_cpickle(
                        'clv_rm_models', config_in['output'], night, lines_label)
                except (FileNotFoundError, IOError):
                    clv_rm_models = load_from_cpickle(
                        'clv_rm_models', config_in['output'], night)
            else:
                # workaround if CLV correction is not available
                clv_rm_models = {'common': {}}
                clv_rm_models['common']['n_radius_grid'] = 3
                clv_rm_models['common']['radius_grid'] = np.asarray(
                    [0.5, 1.0, 1.5])

            processed = {
                'subroutine': subroutine_name,
            }

            processed['common'] = {
                'range': lines_dict['fit_parameters']['range']
            }
            processed['common']['wave'] = np.arange(processed['common']['range'][0],
                                                    processed['common']['range'][1],
                                                    lines_dict['fit_parameters']['bin_step'],
                                                    dtype=np.double)
            processed['common']['size'] = len(processed['common']['wave'])
            processed['common']['step'] = np.ones(
                processed['common']['size'], dtype=np.double) * lines_dict['fit_parameters']['bin_step']

            processed['common_extended'] = {
                'range': lines_dict['range']
            }
            processed['common_extended']['wave'] = np.arange(processed['common_extended']['range'][0],
                                                    processed['common_extended']['range'][1],
                                                    lines_dict['fit_parameters']['bin_step'],
                                                    dtype=np.double)
            processed['common_extended']['size'] = len(
                processed['common_extended']['wave'])
            processed['common_extended']['step'] = np.ones(
                processed['common_extended']['size'], dtype=np.double) * lines_dict['fit_parameters']['bin_step']

            for obs in lists['observations']:

                """ we start from the e2ds file, after correction for blaze and
                    division by the master-out

                Observation data:
                    wave: input_data[obs]['wave']
                    step: input_data[obs]['step']
                    flux: preparation[obs]['ratio']
                    ferr: preparation[obs]['ratio']
                """

                """ First step: we rebin the spectra in the Stellar Reference Frame,
                    with the step size decided by the user specifically for the fit
                """

                if pca_output:
                    preserve_flux = False
                    blaze = np.ones_like(calib_data['blaze'])
                else:
                    preserve_flux = input_data[obs].get('absolute_flux', True)
                    blaze = calib_data['blaze']

                processed[obs] = {}
                processed[obs]['rebinned'] = \
                    rebin_2d_to_1d(input_data[obs]['wave'],
                                    input_data[obs]['step'],
                                    preparation[obs]['ratio'],
                                    blaze,
                                    processed['common']['wave'],
                                    processed['common']['step'],
                                    rv_shift=observational_pams[obs]['rv_shift_ORF2SRF'],
                                    preserve_flux=preserve_flux)
                processed[obs]['rebinned_err'] = \
                    rebin_2d_to_1d(input_data[obs]['wave'],
                                    input_data[obs]['step'],
                                    preparation[obs]['ratio_err'],
                                    blaze,
                                    processed['common']['wave'],
                                    processed['common']['step'],
                                    rv_shift=observational_pams[obs]['rv_shift_ORF2SRF'],
                                    preserve_flux=preserve_flux,
                                    is_error=True)

                processed[obs]['rebinned_extended'] = \
                    rebin_2d_to_1d(input_data[obs]['wave'],
                                    input_data[obs]['step'],
                                    preparation[obs]['ratio'],
                                    blaze,
                                    processed['common_extended']['wave'],
                                    processed['common_extended']['step'],
                                    rv_shift=observational_pams[obs]['rv_shift_ORF2SRF'],
                                    preserve_flux=preserve_flux)



            if norm_pams['normalize_transmission'] and norm_pams['normalization_model'] == 'polynomial':

                """ Continuum normalization preparatory steps:
                    1) exclusion of regions with planetary lines
                    2) exclusion of regions with stellar lines
                    3) Polynomial fit of selected regions
                        Boolean array initialized to all True values, fit is
                        performed on the extended region and then applied to the fit subset
                """
                processed['common_extended']['line_exclusion'] = (
                    processed['common_extended']['wave'] > 0.)

                """ Continuum normalization:
                    1) exclusion of regions with planetary lines, taking into
                        account the planetary RV semi-amplitude
                """
                for line_key, line_val in lines_dict['lines'].items():
                    line_extension = 1.2 * \
                        planet_dict['RV_semiamplitude'][0] * \
                            line_val / speed_of_light_km
                    processed['common_extended']['line_exclusion'] = processed['common_extended']['line_exclusion'] & (
                        np.abs(processed['common_extended']['wave']-line_val) > line_extension)

                """ Continuum normalization:
                    2) exclusion of regions with planetary lines, taking into
                        account the planetary RV semi-amplitude
                """

                try:
                    stellar_spectrum_rebinned = rebin_1d_to_1d(clv_rm_models['common']['wave'],
                                        clv_rm_models['common']['step'],
                                        clv_rm_models['common']['norm_convolved'],
                                        processed['common_extended']['wave'],
                                        processed['common_extended']['step'],
                                        preserve_flux=False)

                    stellar_spectrum_derivative = first_derivative(
                        processed['common_extended']['wave'], stellar_spectrum_rebinned)

                    cont_10perc = np.percentile(np.abs(stellar_spectrum_derivative), norm_pams['percentile_selection'])

                    processed['common_extended']['line_exclusion'] = processed['common_extended']['line_exclusion'] \
                            & (np.abs(stellar_spectrum_derivative) < cont_10perc) \
                            & (stellar_spectrum_rebinned > norm_pams['lower_threshold'])

                except KeyError:
                    print(
                        "No stellar synthetic spectrum from CLV models, some stellar lines may be included transmission normalization  ")

                for obs in lists['observations']:

                    selection = processed['common_extended']['line_exclusion'] & (
                        processed[obs]['rebinned_extended'] > np.std(processed[obs]['rebinned_extended']))

                    processed[obs]['norm_coeff'] = \
                        np.polynomial.chebyshev.chebfit(processed['common_extended']['wave'][selection],
                                                        processed[obs]['rebinned_extended'][selection],
                                                        norm_pams['spectra_poly_degree'])
                    processed[obs]['continuum'] = np.polynomial.chebyshev.chebval(
                        processed['common']['wave'], processed[obs]['norm_coeff'])
                    processed[obs]['normalized'] = processed[obs]['rebinned'] / \
                        processed[obs]['continuum']
                    processed[obs]['normalized_err'] = processed[obs]['rebinned_err'] / \
                        processed[obs]['continuum']

            elif norm_pams['normalize_transmission']  and (
                    norm_pams['normalization_model'] == 'savgol'
                    or norm_pams['normalization_model'] == 'savitzky-golay'):

                print('  Normalization using Savitzky-Golay filter')

                for obs in lists['observations']:

                    if norm_pams['normalize_rebinned']:

                        processed[obs]['continuum'] = savgol_filter(preparation[obs]['rebinned'],
                                                                window_length=norm_pams['window_length'],
                                                                polyorder=norm_pams['polyorder'],
                                                                mode=norm_pams['mode'],
                                                                cval=norm_pams['cval'])


                    else:
                        normalization_model = preparation[obs]['ratio'] * 0.00
                        for order in range(0, observational_pams['n_orders']):

                            normalization_model[order,:] = savgol_filter(preparation[obs]['ratio'][order,:],
                                                                window_length=norm_pams['window_length'],
                                                                polyorder=norm_pams['polyorder'],
                                                                mode=norm_pams['mode'],
                                                                cval=norm_pams['cval'])

                        processed[obs]['continuum'] = \
                            rebin_2d_to_1d(input_data[obs]['wave'],
                                            input_data[obs]['step'],
                                            normalization_model,
                                            blaze,
                                            processed['common']['wave'],
                                            processed['common']['step'],
                                            rv_shift=observational_pams[obs]['rv_shift_ORF2SRF'],
                                            preserve_flux=preserve_flux)


                    processed[obs]['continuum_coeff'] = None
                    processed[obs]['normalized'] = processed[obs]['rebinned'] / \
                        processed[obs]['continuum']
                    processed[obs]['normalized_err'] = processed[obs]['rebinned_err'] / \
                        processed[obs]['continuum']


            else:
                for obs in lists['observations']:

                    processed[obs]['continuum_coeff'] = None
                    processed[obs]['continuum'] = np.ones_like(processed['common']['wave'])
                    processed[obs]['normalized'] = processed[obs]['rebinned'].copy()
                    processed[obs]['normalized_err'] = processed[obs]['rebinned_err'].copy()


            processed['common']['n_obs'] = len(lists['transit_full'])
            processed['common']['n_radius_grid'] = clv_rm_models['common']['n_radius_grid']
            processed['common']['radius_grid'] = clv_rm_models['common']['radius_grid']
            clv_rm_radius = clv_rm_models['common']['radius_grid']

            """ We are moving the values of interest from dictionaries to arrays
                in order to  speed up the MCMC
                1) clv_rm_grid: array with all the CLV models, as a function of the
                    radius of the planet
                2) time_from_transit: BJD_TDB - T0
                3) planet_RVsinusoid: Fractional RV of the planet (K=1) - from a meshgrid
            """
            clv_rm_grid = np.ones([processed['common']['n_radius_grid'],
                                    processed['common']['n_obs'],
                                    processed['common']['size']],
                                   dtype=np.double)

            time_from_transit = np.empty(
                processed['common']['n_obs'], dtype=np.double)

            transmission_spec = np.empty([processed['common']['n_obs'],
                                          processed['common']['size']],
                                         dtype=np.double)
            transmission_spec_err = np.empty([processed['common']['n_obs'],
                                              processed['common']['size']],
                                             dtype=np.double)

            for i_obs, obs in enumerate(lists['transit_full']):

                time_from_transit[i_obs] = observational_pams[obs]['BJD'] - \
                    observational_pams['time_of_transit']
                # planet_RVsinusoid[i_obs] = np.sin(2*np.pi /  planet_dict['period'][0] * time_from_transit[i_obs])

                transmission_spec[i_obs, :] = processed[obs]['normalized']
                transmission_spec_err[i_obs,
                                      :] = processed[obs]['normalized_err']

                if clv_rm_correction is False:
                    continue

                for i_r in range(0, processed['common']['n_radius_grid']):

                    """ CLV Synthetic models are in the Stellar Reference system,
                        so no shift is required """

                    clv_rm_grid[i_r, i_obs, :] = \
                        rebin_1d_to_1d(clv_rm_models['common']['wave'],
                                       clv_rm_models['common']['step'],
                                       clv_rm_models[obs]['clv_rm_model_convolved_normalized'][i_r, :],
                                       processed['common']['wave'],
                                       processed['common']['step'],
                                       preserve_flux=False)



                    # preserve_flux should be True or False?
                    # False if the spectra are already normalized


            #colors_properties, colors_plot, colors_scatter = make_color_array_matplotlib3(
            #    lists, observational_pams)

            #fig = plt.figure(figsize=(12, 6))
            #gs = GridSpec(2, 2, width_ratios=[50, 1])

            #ax1 = plt.subplot(gs[0, 0])
            #ax2 = plt.subplot(gs[1, 0], sharex=ax1, sharey=ax1)
            #cbax1 = plt.subplot(gs[:, 1])

            #i_r = 0
            #for i_obs, obs in enumerate(lists['transit_full']):

            #    ax1.plot(processed['common']['wave'],
            #            clv_rm_grid[i_r, i_obs, :],
            #            color=colors_plot['mBJD'][obs], alpha=0.2)

            #i_r = processed['common']['n_radius_grid']-1
            #for i_obs, obs in enumerate(lists['transit_full']):

            #    ax2.plot(processed['common']['wave'],
            #            clv_rm_grid[i_r, i_obs, :],
            #            color=colors_plot['mBJD'][obs], alpha=0.2)

            #ax1.set_title(
            #    'Night: {0:s} \n CLV+RM correction, convolved and normalized '.format(night))
            #ax2.set_title('Out of transit')
            #ax2.set_xlabel('$\lambda$ [$\AA$]')

            #sm = plt.cm.ScalarMappable(
            #    cmap=colors_properties['cmap'], norm=colors_properties['norm']['mBJD'])
            #sm.set_array([])  # You have to set a dummy-array for this to work...
            #cbar = plt.colorbar(sm, cax=cbax1)
            #cbar.set_label('BJD - 2450000.0')
            #fig.subplots_adjust(wspace=0.05, hspace=0.4)
            #plt.show()

            #quit()


            remove_outliers = (np.abs(transmission_spec - 1.) > 0.5)
            transmission_spec[remove_outliers] = 1.0
            transmission_spec_err[remove_outliers] = 1.0
            wave_meshgrid, time_meshgrid = np.meshgrid(
                processed['common']['wave'], time_from_transit)
            planet_RVsinusoid = np.sin(
                2*np.pi / planet_dict['period'][0] * time_meshgrid)

            if jitter_flag:
                jitter_index = []
                n_jitter = 1
            else:
                jitter_index = None
                n_jitter = 0

            mcmc_data = {
                'observations': lists['transit_full'],
                'common_wave': processed['common']['wave'],
                'common_step': processed['common']['step'],
                'clv_rm_grid': clv_rm_grid,
                'transmission_spec': transmission_spec,
                'transmission_spec_err': transmission_spec_err,
                'wave_meshgrid': wave_meshgrid,
                'time_meshgrid': time_meshgrid,
                'planet_RVsinusoid': planet_RVsinusoid,
                'clv_rm_radius': clv_rm_models['common']['radius_grid'],
                'n_obs': len(lists['transit_full']),
                'n_radius_grid': clv_rm_models['common']['n_radius_grid'],
                'jitter_index': jitter_index,
                'n_jitter': n_jitter
            }

            save_to_cpickle(subroutine_name + '_data', mcmc_data,
                            config_in['output'], night, lines_label, it_string)

            # Forcing memory deallocation
            clv_rm_models = None
            mcmc_data = None


        print()
        print("transmission_binned_mcmc           ")

        try:
            results_dict = load_from_cpickle(subroutine_name+'_'+sampler_name+'_results', config_in['output'], night, lines_label, it_string)
            print("   Transmission MCMC analysis for lines {0:s}, night: {1:s}  already performed".format(
                lines_label, night))

            pams_dict = results_dict['pams_dict']
            chain_med = results_dict['chain_med']
            boundaries = results_dict['boundaries']
            start_average = np.average(results_dict['point_start'], axis=0)
            ndim = results_dict['ndim']
            med_lines_model = results_dict['results']['lines_model']

            if 'derived' in results_dict:
                recompute_derived = False
            else:
                recompute_derived = True
                results_dict['derived'] = {}

            # TODO improve output
            print('   *** sampler output ')

            for key, val in pams_dict.items():
                print('{0:24s}  {1:4d}  {2:12f}   {3:12f}  {4:12f} (15-84 p) ([{5:9f}, {6:9f}]) (start: {7:9f})'.format(key, val,
                                                                                                        chain_med[val,0],
                                                                                                        chain_med[val,2],
                                                                                                        chain_med[val,1],
                                                                                                        boundaries[val, 0],
                                                                                                        boundaries[val, 1],
                                                                                                        start_average[val])
                                                                                                        )
                if recompute_derived and key[-8:]=='contrast':
                    key_name = key[:-8] + 'Rh'
                    sample_size = len(results_dict['flat_chain'][:,val])
                    planet_ratio_sample = np.random.normal(planet_dict['radius_ratio'][0],planet_dict['radius_ratio'][1],size=sample_size)
                    results_dict['derived'][key_name] = {}
                    results_dict['derived'][key_name]['flat_chain'] = np.sqrt(results_dict['flat_chain'][:,val]/planet_ratio_sample**2 + 1.)
                    results_dict['derived'][key_name]['chain_med'] = compute_value_sigma(results_dict['derived'][key_name]['flat_chain'])

                #    R(h) = np.sqrt(1+h/delta)

                #    print(key[-8:], key[:3])

            print('   *** derived output ')
            for key, val in results_dict['derived'].items():
                chain_med = results_dict['derived'][key]['chain_med']
                print('{0:24s}         {1:12f}   {2:12f}  {3:12f} (15-84 p)'.format(key,
                                                                                     chain_med[0],
                                                                                     chain_med[2],
                                                                                     chain_med[1]))



            continue



        except FileNotFoundError:
            print()

        # getting fit parameters
        lines_center, pams_dict, pams_list, boundaries, theta_start = define_theta_array(
            model_case, lines_dict, planet_dict, n_jitter, allow_emission=allow_emission)
        ndim = len(theta_start)

        if pyde_flag:
            ngen = sampler_pams.get('n_gen', 64000)
        else:
            ngen = 0
        nwalkers_mult = sampler_pams.get('n_walkers_mult', 2)
        nwalkers = sampler_pams.get('n_walkers', nwalkers_mult * ndim)
        nthin = sampler_pams.get('n_thin', 50)
        nsteps = sampler_pams.get('n_steps', 20000)
        nburnin = sampler_pams.get('n_burnin', 10000)
        ncpus = sampler_pams.get('n_cpus', 4)
        ndata = np.size(wave_meshgrid)

        if pams_dict.get('rp_factor', False):
            pam_id = pams_dict['rp_factor']
            boundaries[pam_id, :] = [clv_rm_radius[0], clv_rm_radius[-1]]


        print()
        print('      PyDE + emcee parameters')
        print('         n_dim:                        {0:9.0f}'.format(ndim))
        if pyde_flag:
            print('         n_gen:      (default: 64000)  {0:9.0f}'.format(ngen))
        else:
            print('         no PyDE optimization, MCMC will start from default values')
        print(
            '         n_walkers:  (default: 2*ndim) {0:9.0f}'.format(nwalkers))
        print('         n_steps:    (default: 20000)  {0:9.0f}'.format(nsteps))
        print(
            '         n_burnin:   (default: 10000)  {0:9.0f}'.format(nburnin))
        print('         n_thin:     (default: 50)     {0:9.0f}'.format(nthin))

        population, sampler_chain, sampler_lnprobability, point_start = emcee_lines_fit_functions(
            model_case,
            wave_meshgrid,
            transmission_spec,
            transmission_spec_err,
            clv_rm_radius,
            clv_rm_grid,
            planet_RVsinusoid,
            lines_center,
            jitter_index,
            prior_dict,
            theta_start, boundaries, ndim, nwalkers, ngen, nsteps, nthin, ncpus)

        flat_chain, flat_lnprob, chain_med, chain_MAP, lnprob_med, lnprob_MAP = \
            emcee_flatten_median(population, sampler_chain,
                                 sampler_lnprobability,  nburnin, nthin, nwalkers)
        emcee_compute_BIC_AIC(lnprob_med, lnprob_MAP, ndata, ndim)

        med_lines_model, med_clv_model, med_lines_array, med_planet_K, med_planet_R, med_jitter = \
            return_model(model_case,
                         chain_med[:, 0],
                         wave_meshgrid,
                         clv_rm_radius,
                         clv_rm_grid,
                         planet_RVsinusoid,
                         lines_center,
                         jitter_index)

        map_lines_model, map_clv_model, map_lines_array, map_planet_K, map_planet_R, map_jitter = \
            return_model(model_case,
                         chain_MAP,
                         wave_meshgrid,
                         clv_rm_radius,
                         clv_rm_grid,
                         planet_RVsinusoid,
                         lines_center,
                         jitter_index)

        results_dict = {
            'sampler_name': sampler_name,
            'ndim': ndim,
            'nwalkers': nwalkers,
            'nthin': nthin,
            'nsteps': nsteps,
            'nburnin': nburnin,
            'ndata': ndata,
            'pams_dict': pams_dict,
            'population': population,
            'sampler_chain': sampler_chain,
            'sampler_lnprobability': sampler_lnprobability,
            'theta_start': theta_start,
            'boundaries': boundaries,
            'flat_chain': flat_chain,
            'flat_lnprob': flat_lnprob,
            'chain_med': chain_med,
            'chain_MAP': chain_MAP,
            'lnprob_med': lnprob_med,
            'lnprob_MAP': lnprob_MAP,
            'lines_center': lines_center,
            'point_start': point_start,
            'theta_start': theta_start,
        }

        results_dict['results'] = {
            'lines_model': med_lines_model,
            'clv_model': med_clv_model,
            'lines_array': med_lines_array,
            'planet_K': med_planet_K,
            'planet_R': med_planet_R,
            'jitter': med_jitter
        }

        results_dict['results_MAP'] = {
            'lines_model': map_lines_model,
            'clv_model': map_clv_model,
            'lines_array': map_lines_array,
            'planet_K': map_planet_K,
            'planet_R': map_planet_R,
            'jitter': map_jitter
        }

        results_dict['results']['observational_pams'] = {}
        results_dict['results_MAP']['observational_pams'] = {}

        for obs in lists['observations']:

            results_dict['results']['observational_pams'][obs] = {}
            results_dict['results_MAP']['observational_pams'][obs] = {}

            """ RV shift from the observer RF to the planet RF
                STRONG ASSUMPTIONS:
                    - there is only the transiting planet in the system
                    - the planet has null eccentricity
                    - linear approximation or the orbit near the transit event

                Computation is performed by moving to the Solar Barycenter, than to the Stellar System Barycenter
                and finally onto the planet
            """
            results_dict['results']['observational_pams'][obs]['rv_shift_ORF2PRF'] = \
                observational_pams[obs]['BERV'] \
                - observational_pams['RV_star']['RV_systemic'] \
                - results_dict['results']['planet_K'] \
                * (observational_pams[obs]['BJD'] - observational_pams['time_of_transit']) \
                / planet_dict['period'][0] * 2 * np.pi

            results_dict['results_MAP']['observational_pams'][obs]['rv_shift_ORF2PRF'] = \
                observational_pams[obs]['BERV'] \
                - observational_pams['RV_star']['RV_systemic'] \
                - results_dict['results_MAP']['planet_K'] \
                * (observational_pams[obs]['BJD'] - observational_pams['time_of_transit']) \
                / planet_dict['period'][0] * 2 * np.pi

            """ RV shift from Stellar Rest Frame to Planetary Rest Frame
                We have to take into account the RV of star relatively to the Barycenter
            """
            results_dict['results']['observational_pams'][obs]['rv_shift_SRF2PRF'] = \
                + observational_pams[obs]['RV_bjdshift'] \
                - results_dict['results']['planet_K'] \
                * (observational_pams[obs]['BJD'] - observational_pams['time_of_transit']) \
                / planet_dict['period'][0] * 2 * np.pi

            results_dict['results_MAP']['observational_pams'][obs]['rv_shift_SRF2PRF'] = \
                + observational_pams[obs]['RV_bjdshift'] \
                - results_dict['results_MAP']['planet_K'] \
                * (observational_pams[obs]['BJD'] - observational_pams['time_of_transit']) \
                / planet_dict['period'][0] * 2 * np.pi


        results_dict['derived'] = {}

        # TODO improve output
        print('   *** sampler output ')

        start_average = np.average(results_dict['point_start'], axis=0)
        for key, val in pams_dict.items():
            print('{0:24s}  {1:4d}  {2:12f}   {3:12f}  {4:12f} (15-84 p) ([{5:9f}, {6:9f}]) (start: {7:9f})'.format(key, val,
                                                                                                    chain_med[val,0],
                                                                                                    chain_med[val,2],
                                                                                                    chain_med[val,1],
                                                                                                    boundaries[val, 0],
                                                                                                    boundaries[val, 1],
                                                                                                    start_average[val])
                                                                                                    )


            if key[-8:]=='contrast':
                key_name = key[:-8] + 'Rh'
                sample_size = len(results_dict['flat_chain'][:,val])
                planet_ratio_sample = np.random.normal(planet_dict['radius_ratio'][0],planet_dict['radius_ratio'][1],size=sample_size)
                results_dict['derived'][key_name] = {}
                results_dict['derived'][key_name]['flat_chain'] = np.sqrt(results_dict['flat_chain'][:,val]/planet_ratio_sample**2 + 1.)
                results_dict['derived'][key_name]['chain_med'] = compute_value_sigma(results_dict['derived'][key_name]['flat_chain'])


        print('   *** derived output ')
        for key, val in results_dict['derived'].items():
            chain_med = results_dict['derived'][key]['chain_med']
            print('{0:24s}         {1:12f}   {2:12f}  {3:12f} (15-84 p)'.format(key,
                                                                                    chain_med[0],
                                                                                    chain_med[2],
                                                                                    chain_med[1]))


        save_to_cpickle(subroutine_name+'_'+sampler_name+'_results',
                        results_dict, config_in['output'], night, lines_label, it_string)

        # print('   *** physical output')
        #
        #        results_dict['results'] = {
        #    'lines_model': med_lines_model,
        #    'clv_model': med_clv_model,
        #    'lines_array': med_lines_array,
        #    'planet_K': med_planet_K,
        #    'planet_R': med_planet_R,
        #    'jitter': med_jitter
        # }

    """ Analysis of the entire dataset """

    print()
    try:
        all_mcmc_data = load_from_cpickle(subroutine_name+'_data', config_in['output'], night='', lines=lines_label, it_string=it_string)

        all_clv_rm_radius = all_mcmc_data['clv_rm_radius']
        all_clv_rm_grid = all_mcmc_data['clv_rm_grid']
        all_transmission_spec = all_mcmc_data['transmission_spec']
        all_transmission_spec_err = all_mcmc_data['transmission_spec_err']
        all_wave_meshgrid = all_mcmc_data['wave_meshgrid']
        all_time_meshgrid = all_mcmc_data['time_meshgrid']
        all_planet_RVsinusoid = all_mcmc_data['planet_RVsinusoid']
        all_observations = all_mcmc_data['observations']
        all_n_obs = all_mcmc_data['n_obs']
        all_n_radius_grid = all_mcmc_data['n_radius_grid']
        all_jitter_index = all_mcmc_data['jitter_index']
        n_jitter = all_mcmc_data['n_jitter']

    except:

        n_jitter = 0

        for night in night_dict:

            mcmc_data = load_from_cpickle(subroutine_name+'_data', config_in['output'], night, lines_label, it_string=it_string)

            try:
                # Building the arrays for the full analysis

                all_clv_rm_grid = np.concatenate(
                    (all_clv_rm_grid, mcmc_data['clv_rm_grid']), axis=1)
                all_transmission_spec = np.concatenate(
                    (all_transmission_spec, mcmc_data['transmission_spec']))
                all_transmission_spec_err = np.concatenate(
                    (all_transmission_spec_err, mcmc_data['transmission_spec_err']))
                all_wave_meshgrid = np.concatenate(
                    (all_wave_meshgrid, mcmc_data['wave_meshgrid']))
                all_time_meshgrid = np.concatenate(
                    (all_time_meshgrid, mcmc_data['time_meshgrid']))
                all_planet_RVsinusoid = np.concatenate(
                    (all_planet_RVsinusoid, mcmc_data['planet_RVsinusoid']))
                all_observations = np.concatenate(
                    (all_observations, mcmc_data['observations']))
                all_n_obs += mcmc_data['n_obs']

                if jitter_flag:
                    all_jitter_index = np.concatenate(
                        (all_jitter_index, n_jitter*np.ones(np.shape(mcmc_data['wave_meshgrid']), dtype=np.int16)))
                    n_jitter += 1

            except NameError:
                """ This error is expected when retrieving the data of the first night"""
                all_clv_rm_radius = mcmc_data['clv_rm_radius']
                all_clv_rm_grid = mcmc_data['clv_rm_grid']
                all_transmission_spec = mcmc_data['transmission_spec']
                all_transmission_spec_err = mcmc_data['transmission_spec_err']
                all_wave_meshgrid = mcmc_data['wave_meshgrid']
                all_time_meshgrid = mcmc_data['time_meshgrid']
                all_planet_RVsinusoid = mcmc_data['planet_RVsinusoid']
                all_observations = mcmc_data['observations']
                all_n_obs = mcmc_data['n_obs']
                all_n_radius_grid = mcmc_data['n_radius_grid']

                if jitter_flag:
                    all_jitter_index = n_jitter * \
                        np.ones(
                            np.shape(mcmc_data['wave_meshgrid']), dtype=np.int16)
                    n_jitter += 1
                else:
                    all_jitter_index = None

        all_mcmc_data = {
            'observations': all_observations,
            'clv_rm_grid': all_clv_rm_grid,
            'transmission_spec': all_transmission_spec,
            'transmission_spec_err': all_transmission_spec_err,
            'wave_meshgrid': all_wave_meshgrid,
            'time_meshgrid': all_time_meshgrid,
            'planet_RVsinusoid': all_planet_RVsinusoid,
            'clv_rm_radius': all_clv_rm_radius,
            'n_obs': all_n_obs,
            'n_radius_grid': all_n_radius_grid,
            'jitter_index': all_jitter_index,
            'n_jitter': n_jitter
        }

        save_to_cpickle(subroutine_name+'_data', all_mcmc_data,
                        config_in['output'], night='', lines=lines_label, it_string=it_string)

    try:
        results_dict = load_from_cpickle(subroutine_name+ '_'+ sampler_name+'_results',
                                         config_in['output'], night='', lines=lines_label, it_string=it_string)
        print("   Transmission MCMC analysis for lines {0:s} already performed ".format(
                lines_label))

        pams_dict = results_dict['pams_dict']
        chain_med = results_dict['chain_med']
        boundaries = results_dict['boundaries']
        ndim = results_dict['ndim']
        start_average = np.average(results_dict['point_start'], axis=0)

        if 'derived' in results_dict:
            recompute_derived = False
        else:
            recompute_derived = True
            results_dict['derived'] = {}

        # TODO improve output
        print('   *** sampler output ')

        for key, val in pams_dict.items():
            print('{0:24s}  {1:4d}  {2:12f}   {3:12f}  {4:12f} (15-84 p) ([{5:9f}, {6:9f}]) (start: {7:9f})'.format(key, val,
                                                                                                    chain_med[val,0],
                                                                                                    chain_med[val,2],
                                                                                                    chain_med[val,1],
                                                                                                    boundaries[val, 0],
                                                                                                    boundaries[val, 1],
                                                                                                    start_average[val])
                                                                                                    )

            if recompute_derived and key[-8:]=='contrast':
                key_name = key[:-8] + 'Rh'
                sample_size = len(results_dict['flat_chain'][:,val])
                planet_ratio_sample = np.random.normal(planet_dict['radius_ratio'][0],planet_dict['radius_ratio'][1],size=sample_size)
                results_dict['derived'][key_name] = {}
                results_dict['derived'][key_name]['flat_chain'] = np.sqrt(results_dict['flat_chain'][:,val]/planet_ratio_sample**2 + 1.)
                results_dict['derived'][key_name]['chain_med'] = compute_value_sigma(results_dict['derived'][key_name]['flat_chain'])
            #    R(h) = np.sqrt(1+h/delta)

                #    print(key[-8:], key[:3])

        print('   *** derived output ')
        for key, val in results_dict['derived'].items():
            chain_med = results_dict['derived'][key]['chain_med']
            print('{0:24s}         {1:12f}   {2:12f}  {3:12f} (15-84 p)'.format(key,
                                                                                    chain_med[0],
                                                                                    chain_med[2],
                                                                                    chain_med[1]))





    except FileNotFoundError:

        lines_center, pams_dict, pams_list, boundaries, theta_start = define_theta_array(
            model_case, lines_dict, planet_dict, n_jitter, allow_emission=allow_emission)
        ndim = len(theta_start)
        ngen = sampler_pams.get('n_gen', 64000)
        nwalkers_mult = sampler_pams.get('n_walkers_mult', 2)
        nwalkers = sampler_pams.get('n_walkers', nwalkers_mult * ndim)
        nthin = sampler_pams.get('n_thin', 50)
        nsteps = sampler_pams.get('n_steps', 20000)
        nburnin = sampler_pams.get('n_burnin', 10000)
        ncpus = sampler_pams.get('n_cpus', 4)
        ndata = np.size(all_wave_meshgrid)

        if pams_dict.get('rp_factor', False):
            pam_id = pams_dict['rp_factor']
            boundaries[pam_id, :] = [clv_rm_radius[0], clv_rm_radius[-1]]

        print()
        print('      PyDE + emcee parameters')
        print('         n_dim:                        {0:9.0f}'.format(ndim))
        print(
            '         n_walkers:  (default: 2*ndim) {0:9.0f}'.format(nwalkers))
        print('         n_gen:      (default: 64000)  {0:9.0f}'.format(ngen))
        print('         n_steps:    (default: 20000)  {0:9.0f}'.format(nsteps))
        print(
            '         n_burnin:   (default: 10000)  {0:9.0f}'.format(nburnin))
        print('         n_thin:     (default: 50)     {0:9.0f}'.format(nthin))

        population, sampler_chain, sampler_lnprobability, point_start = emcee_lines_fit_functions(
            model_case,
            all_wave_meshgrid,
            all_transmission_spec,
            all_transmission_spec_err,
            all_clv_rm_radius,
            all_clv_rm_grid,
            all_planet_RVsinusoid,
            lines_center,
            all_jitter_index,
            prior_dict,
            theta_start, boundaries, ndim, nwalkers, ngen, nsteps, nthin, ncpus)

        flat_chain, flat_lnprob, chain_med, chain_MAP, lnprob_med, lnprob_MAP = \
            emcee_flatten_median(population, sampler_chain,
                                 sampler_lnprobability,  nburnin, nthin, nwalkers)
        emcee_compute_BIC_AIC(lnprob_med, lnprob_MAP, ndata, ndim)

        med_lines_model, med_clv_model, med_lines_array, med_planet_K, med_planet_R, med_jitter = \
            return_model(model_case,
                         chain_med[:, 0],
                         all_wave_meshgrid,
                         all_clv_rm_radius,
                         all_clv_rm_grid,
                         all_planet_RVsinusoid,
                         lines_center,
                         all_jitter_index)

        map_lines_model, map_clv_model, map_lines_array, map_planet_K, map_planet_R, map_jitter = \
            return_model(model_case,
                         chain_MAP,
                         all_wave_meshgrid,
                         all_clv_rm_radius,
                         all_clv_rm_grid,
                         all_planet_RVsinusoid,
                         lines_center,
                         all_jitter_index)

        results_dict = {
            'sampler_name': sampler_name,
            'ndim': ndim,
            'nwalkers': nwalkers,
            'nthin': nthin,
            'nsteps': nsteps,
            'nburnin': nburnin,
            'ndata': ndata,
            'pams_dict': pams_dict,
            'population': population,
            'sampler_chain': sampler_chain,
            'sampler_lnprobability': sampler_lnprobability,
            'theta_start': theta_start,
            'boundaries': boundaries,
            'flat_chain': flat_chain,
            'flat_lnprob': flat_lnprob,
            'chain_med': chain_med,
            'chain_MAP': chain_MAP,
            'lnprob_med': lnprob_med,
            'lnprob_MAP': lnprob_MAP,
            'lines_center': lines_center,
            'point_start': point_start,
            'theta_start': theta_start
            #'BIC': BIC,
            #'BIC_map': BIC_map
        }

        results_dict['results'] = {
            'lines_model': med_lines_model,
            'clv_model': med_clv_model,
            'lines_array': med_lines_array,
            'planet_K': med_planet_K,
            'planet_R': med_planet_R,
            'jitter': med_jitter
        }


        results_dict['results_MAP'] = {
            'lines_model': map_lines_model,
            'clv_model': map_clv_model,
            'lines_array': map_lines_array,
            'planet_K': map_planet_K,
            'planet_R': map_planet_R,
            'jitter': map_jitter
        }

        results_dict['results']['observational_pams'] = {}
        results_dict['results_MAP']['observational_pams'] = {}

        for night in night_dict:

            lists = load_from_cpickle('lists', config_in['output'], night)
            observational_pams = load_from_cpickle(
                'observational_pams', config_in['output'], night)

            """ No differentiation by night """
            for obs in lists['observations']:

                results_dict['results']['observational_pams'][obs] = {}
                results_dict['results_MAP']['observational_pams'][obs] = {}

                """ RV shift from the observer RF to the planet RF
                    STRONG ASSUMPTIONS:
                        - there is only the transiting planet in the system
                        - the planet has null eccentricity
                        - linear approximation or the orbit near the transit event

                    Computation is performed by moving to the Solar Barycenter, than to the Stellar System Barycenter
                    and finally onto the planet
                """
                results_dict['results']['observational_pams'][obs]['rv_shift_ORF2PRF'] = \
                    observational_pams[obs]['BERV'] \
                    - observational_pams['RV_star']['RV_systemic'] \
                    - results_dict['results']['planet_K'] \
                    * (observational_pams[obs]['BJD'] - observational_pams['time_of_transit']) \
                    / planet_dict['period'][0] * 2 * np.pi

                results_dict['results_MAP']['observational_pams'][obs]['rv_shift_ORF2PRF'] = \
                    observational_pams[obs]['BERV'] \
                    - observational_pams['RV_star']['RV_systemic'] \
                    - results_dict['results_MAP']['planet_K'] \
                    * (observational_pams[obs]['BJD'] - observational_pams['time_of_transit']) \
                    / planet_dict['period'][0] * 2 * np.pi

                """ RV shift from Stellar Rest Frame to Planetary Rest Frame
                    We have to take into account the RV of star relatively to the Barycenter
                """
                results_dict['results']['observational_pams'][obs]['rv_shift_SRF2PRF'] = \
                    + observational_pams[obs]['RV_bjdshift'] \
                    - results_dict['results']['planet_K'] \
                    * (observational_pams[obs]['BJD'] - observational_pams['time_of_transit']) \
                    / planet_dict['period'][0] * 2 * np.pi

                results_dict['results_MAP']['observational_pams'][obs]['rv_shift_SRF2PRF'] = \
                    + observational_pams[obs]['RV_bjdshift'] \
                    - results_dict['results_MAP']['planet_K'] \
                    * (observational_pams[obs]['BJD'] - observational_pams['time_of_transit']) \
                    / planet_dict['period'][0] * 2 * np.pi



        start_average = np.average(results_dict['point_start'], axis=0)

        results_dict['derived'] = {}

        # TODO improve output
        print('   *** sampler output ')
        for key, val in pams_dict.items():
            print('{0:24s}  {1:4d}  {2:12f}   {3:12f}  {4:12f} (15-84 p) ([{5:9f}, {6:9f}]) (start: {7:9f})'.format(key, val,
                                                                                                    chain_med[val,0],
                                                                                                    chain_med[val,2],
                                                                                                    chain_med[val,1],
                                                                                                    boundaries[val, 0],
                                                                                                    boundaries[val, 1],
                                                                                                    start_average[val])
                                                                                                    )
            if key[-8:]=='contrast':
                key_name = key[:-8] + 'Rh'
                sample_size = len(results_dict['flat_chain'][:,val])
                print(sample_size)
                planet_ratio_sample = np.random.normal(planet_dict['radius_ratio'][0],planet_dict['radius_ratio'][1],size=sample_size)
                results_dict['derived'][key_name] = {}
                results_dict['derived'][key_name]['flat_chain'] = np.sqrt(results_dict['flat_chain'][:,val]/planet_ratio_sample**2 + 1.)
                results_dict['derived'][key_name]['chain_med'] = compute_value_sigma(results_dict['derived'][key_name]['flat_chain'])


        print('   *** derived output ')
        for key, val in results_dict['derived'].items():
            chain_med = results_dict['derived'][key]['chain_med']
            print('{0:24s}         {1:12f}   {2:12f}  {3:12f} (15-84 p)'.format(key,
                                                                                    chain_med[0],
                                                                                    chain_med[2],
                                                                                    chain_med[1]))

        save_to_cpickle(subroutine_name +'_'+sampler_name+'_results',
                        results_dict, config_in['output'], night='', lines=lines_label, it_string=it_string)


        print('MCMC completed')

    # Update planet parameters
    # deprecated
    # try:
    #    _ = load_from_cpickle(
    #        'observational', config_in['output'], night, lines_label)
    #    print("   Transmission MCMC results for lines {0:s} already store in observational array".format(
    #            lines_label))
    # except FileNotFoundError:
    #
    #    results_full = load_from_cpickle('transmission_mcmc_'+sampler_name+'_results',
    #                                     config_in['output'], night='', lines=lines_label)
    #
    #    for night in night_dict:
    #
    #        results_night = load_from_cpickle('transmission_mcmc_'+sampler_name+'_results',
    #                                     config_in['output'], night=night, lines=lines_label)


    #        lists = load_from_cpickle('lists', config_in['output'], night)
    #        observational_pams = load_from_cpickle(
    #            'observational_pams', config_in['output'], night)



    #        for obs in lists['observations']:
    #
    #            """ RV shift from the observer RF to the planet RF
    #                STRONG ASSUMPTIONS:
    #                    - there is only the transiting planet in the system
    #                    - the planet has null eccentricity
    #                    - linear approximation or the orbit near the transit event
    #
    #                Computation is performed by moving to the Solar Barycenter, than to the Stellar System Barycenter
    #                and finally onto the planet
    #            """
    #            observational_pams[obs]['rv_shift_ORF2PRF'] = \
    #                observational_pams[obs]['BERV'] \
    #                - observational_pams['RV_star']['RV_systemic'] \
    #                - results_full['results']['planet_K'] \
    #                * (observational_pams[obs]['BJD'] - observational_pams['time_of_transit']) \
    #                / planet_dict['period'][0] * 2 * np.pi

    #            """ RV shift from Stellar Rest Frame to Planetary Rest Frame
    #                We have to take into account the RV of star relatively to the Barycenter
    #            """
    #            observational_pams[obs]['rv_shift_SRF2PRF'] = \
    #                + observational_pams[obs]['RV_bjdshift'] \
    #                - results_full['results']['planet_K'] \
    #                * (observational_pams[obs]['BJD'] - observational_pams['time_of_transit']) \
    #                / planet_dict['period'][0] * 2 * np.pi

    #        observational_pams['Rp_factor'] = results_full['results']['planet_R']
    #        observational_pams['lines_array'] = results_full['results']['lines_array']
    #        observational_pams['jitter'] = results_full['results']['jitter']

    #        save_to_cpickle('observational', observational_pams,
    #                        config_in['output'], night, lines_label)



def plot_transmission_binned_mcmc(config_in, lines_label, night_input='', reference='planetRF', pca_iteration=-1):

    night_dict = from_config_get_nights(config_in)
    planet_dict = from_config_get_planet(config_in)
    star_dict = from_config_get_star(config_in)
    clv_rm_dict = from_config_get_clv_rm(config_in)

    spectral_lines = from_config_get_spectral_lines(config_in)
    lines_dict = spectral_lines[lines_label]
    sampler_pams = lines_dict['sampler_parameters']
    sampler_name = sampler_pams.get('sampler_name', 'emcee')

    if night_input == '':
        night_list = ['']
    else:
        night_list = np.atleast_1d(night_input)

    os.system('mkdir -p plots')


    # Workaround to check if the transmission spectrum has been obtained through PCA iterations
    for night in night_dict:
        preparation_input = load_from_cpickle('transmission_preparation', config_in['output'], night)

        if preparation_input.get('pca_output', False):
            if pca_iteration >= 0:
                it_string = str(pca_iteration).zfill(2)
            else:
                it_string = str(preparation_input.get('ref_iteration')).zfill(2)
        else:
            it_string = ''
        preparation_input = None
        break


    for night in night_list:

        results_dict = load_from_cpickle(subroutine_name+'_'+sampler_name+'_results', config_in['output'], night, lines_label, it_string)
        print("   Transmission MCMC analysis for lines {0:s}, night: {1:s}  already performed".format(
            lines_label, night))

        if night == '':
            chains_dir = 'plots/mcmc_binned_chains_full/'
        else:
            chains_dir = 'plots/mcmc_binned_chains_' + night + '/'
        os.system('mkdir -p ' + chains_dir)


        pams_dict = results_dict['pams_dict']
        chain_med = results_dict['chain_med']
        lnprob_med = results_dict['lnprob_med']
        boundaries = results_dict['boundaries']
        flat_chain = results_dict['flat_chain']
        flat_lnprob = results_dict['flat_lnprob']
        nthin = results_dict['nthin']
        nsteps = results_dict['nsteps']
        nburnin = results_dict['nburnin']

        sampler_chain = results_dict['sampler_chain']

        start_average = np.average(results_dict['point_start'], axis=0)
        ndim = results_dict['ndim']
        med_lines_model = results_dict['results']['lines_model']

        if 'derived' in results_dict:
            recompute_derived = False
        else:
            recompute_derived = True
            results_dict['derived'] = {}

        # TODO improve output
        print('   *** sampler output (plotting the chains)')

        sample_size = np.size(flat_chain, axis=0)
        dimen_size = np.size(flat_chain, axis=1)
        corner_plot = {
            'samples': np.zeros([sample_size, dimen_size + 1]),
            'labels': [],
            'truths': [],
            'start': [],
        }

        i_corner = 0


        for key, val in pams_dict.items():
            print('{0:24s}  {1:4d}  {2:12f}   {3:12f}  {4:12f} (15-84 p) ([{5:9f}, {6:9f}]) (start: {7:9f})'.format(key, val,
                                                                                                    chain_med[val,0],
                                                                                                    chain_med[val,2],
                                                                                                    chain_med[val,1],
                                                                                                    boundaries[val, 0],
                                                                                                    boundaries[val, 1],
                                                                                                    start_average[val])
                                                                                                    )
            if recompute_derived and key[-8:]=='contrast':
                key_name = key[:-8] + 'Rh'
                planet_ratio_sample = np.random.normal(planet_dict['radius_ratio'][0],planet_dict['radius_ratio'][1],size=sample_size)
                results_dict['derived'][key_name] = {}
                results_dict['derived'][key_name]['flat_chain'] = np.sqrt(results_dict['flat_chain'][:,val]/planet_ratio_sample**2 + 1.)
                results_dict['derived'][key_name]['chain_med'] = compute_value_sigma(results_dict['derived'][key_name]['flat_chain'])


            corner_plot['samples'][:, i_corner] = flat_chain[:, val]
            corner_plot['labels'].append(re.sub('_', ' ', key))
            corner_plot['truths'].append(chain_med[val, 0])
            corner_plot['start'].append(start_average[val])
            i_corner += 1


            file_name = chains_dir + repr(val) + '.png'
            fig = plt.figure(figsize=(12, 12))
            plt.title(key)
            plt.plot(sampler_chain[:, :, val].T, '-', alpha=0.5)
            plt.axvline(nburnin / nthin, c='r')
            plt.savefig(file_name, bbox_inches='tight', dpi=300)
            plt.close(fig)



        corner_plot['samples'][:, -1] = flat_lnprob[:]
        corner_plot['labels'].append('ln-prob')
        corner_plot['truths'].append(lnprob_med[0])
        corner_plot['start'].append(None)

            #    R(h) = np.sqrt(1+h/delta)

            #    print(key[-8:], key[:3])

        print('   *** derived output ')
        for key, val in results_dict['derived'].items():
            chain_med = results_dict['derived'][key]['chain_med']
            print('{0:24s}         {1:12f}   {2:12f}  {3:12f} (15-84 p)'.format(key,
                                                                                    chain_med[0],
                                                                                    chain_med[2],
                                                                                    chain_med[1]))
        print('   *** corner plot using pyGTC output ')

        filename_rad = subroutine_name + '_' + reference + '_cornerplot'
        output_file = get_filename(filename_rad, config_in['output'], night=night, lines=lines_label, it_string=it_string, extension='.pdf')
        print('   *** filename: ', output_file)

        try:
            GTC = pygtc.plotGTC(chains=corner_plot['samples'],
                                paramNames=corner_plot['labels'],
                                truths=[corner_plot['truths'],corner_plot['start']],
                                GaussianConfLevels=True,
                                nConfidenceLevels=3,
                                figureSize=12,
                                labelRotation= (True,True),
                                plotName='plots/'+output_file)
        except:
            GTC = pygtc.plotGTC(chains=corner_plot['samples'],
                                paramNames=corner_plot['labels'],
                                truths=[corner_plot['truths'],corner_plot['start']],
                                GaussianConfLevels=True,
                                nConfidenceLevels=2,
                                figureSize=12,
                                labelRotation= (True,True),
                                plotName='plots/'+output_file)


        GTC = None


        continue


def plot_transmission_binned_mcmc_deprecated(config_in, lines_label, night_input=''):
    night_dict = from_config_get_nights(config_in)

    spectral_lines = from_config_get_spectral_lines(config_in)
    lines_dict = spectral_lines[lines_label]

    if night_input == '':
        night_list = night_dict
    else:
        night_list = np.atleast_1d(night_input)

    for night in night_list:

        """ Retrieving the list of observations"""
        lists = load_from_cpickle('lists', config_in['output'], night)

        """ Retrieving the analysis"""
        try:
            clv_rm_corrected = load_from_cpickle(subroutine_name, config_in['output'], night)

            mcmc_data = load_from_cpickle(subroutine_name + '_data', config_in['output'], night, lines_label)

            clv_rm_radius = mcmc_data['clv_rm_radius']
            clv_rm_grid = mcmc_data['clv_rm_grid']
            transmission_spec = mcmc_data['transmission_spec']
            transmission_spec_err = mcmc_data['transmission_spec_err']
            wave_meshgrid = mcmc_data['wave_meshgrid']
            time_meshgrid = mcmc_data['time_meshgrid']
            planet_RVsinusoid = mcmc_data['planet_RVsinusoid']

        except:
            print("No transmission spectrum results, no plots")
            print()
            continue

        """ Creation of the color array, based on the BJD of the observations
        """
        bjd = []
        am = []

        for obs in lists['observations']:
            bjd.append(clv_rm_corrected[obs]['BJD'] - 2450000.0)
            am.append(clv_rm_corrected[obs]['AIRMASS'])



        from matplotlib.colors import BoundaryNorm
        from matplotlib.ticker import MaxNLocator
        cmap = plt.get_cmap('coolwarm')

        plot_data = transmission_spec.copy()


        from SLOPpy.subroutines.math_functions import interpolate2d_grid_nocheck
        plot_data = interpolate2d_grid_nocheck(1.000, clv_rm_radius, clv_rm_grid)
        #clv_model = interpolate2d_grid_nocheck(1.000, clv_rm_radius, clv_rm_grid)
        vmin = plot_data.min()
        vmax =  plot_data.max()

        levels = MaxNLocator(nbins=15).tick_values(vmin, vmax)
        norm = BoundaryNorm(levels, ncolors=cmap.N, clip=True)
        plt.figure(figsize=(15, 10))

        PCF = plt.contourf(wave_meshgrid, time_meshgrid, plot_data, levels=levels, cmap=cmap)
        cbar = plt.colorbar(PCF)
        cbar.ax.set_ylabel('Intensity')
        plt.show()

        levels = MaxNLocator(nbins=15).tick_values(vmin, vmax)
        norm = BoundaryNorm(levels, ncolors=cmap.N, clip=True)
        plt.figure(figsize=(15, 10))

        PCM = plt.pcolormesh(wave_meshgrid, time_meshgrid, plot_data,
                             vmin=vmin, vmax=vmax, cmap=cmap)
        cbar = plt.colorbar(PCM)
        cbar.ax.set_ylabel('Intensity')
        plt.show()




        plot_data = transmission_spec.copy()


        plot_data = transmission_spec.copy()
        #clv_model = interpolate2d_grid_nocheck(1.000, clv_rm_radius, clv_rm_grid)
        vmin = plot_data.min()
        vmax =  plot_data.max()

        levels = MaxNLocator(nbins=15).tick_values(vmin, vmax)
        norm = BoundaryNorm(levels, ncolors=cmap.N, clip=True)
        plt.figure(figsize=(15, 10))

        PCF = plt.contourf(wave_meshgrid, time_meshgrid, plot_data, levels=levels, cmap=cmap)
        cbar = plt.colorbar(PCF)
        cbar.ax.set_ylabel('Intensity')
        plt.show()

        levels = MaxNLocator(nbins=15).tick_values(vmin, vmax)
        norm = BoundaryNorm(levels, ncolors=cmap.N, clip=True)
        plt.figure(figsize=(15, 10))

        PCM = plt.pcolormesh(wave_meshgrid, time_meshgrid, plot_data,
                             vmin=vmin, vmax=vmax, cmap=cmap)
        cbar = plt.colorbar(PCM)
        cbar.ax.set_ylabel('Intensity')
        plt.show()



from __future__ import print_function, division
import numpy as np
from astropy.io import fits
from SLOPpy.subroutines.rebin_subroutines import *
from SLOPpy.subroutines.constants import *

""" Empty module left here for back-compatibility, as almost every module 
    is importing this one rather than the rebin_subroutines
   Everything has been moved into the instruments folder for 
   easier handling of different instruments
"""


#from SLOPpy.instruments.HARPN_DRSv3 import *
#from SLOPpy.instruments.HARPS_DRSv3 import *
#from SLOPpy.instruments.PEPSI_reduced import *
#
#def get_calib_data(instrument, archive, file_rad, fiber='A', order_selection=None):
#
#    if instrument =='HARPS-N':
#        return HARPN_DRSv3_get_calib_data(archive, file_rad, fiber=fiber, order_selection=order_selection)
#    elif instrument =='HARPS':
#        return HARPS_DRSv3_get_calib_data(archive, file_rad, fiber=fiber, order_selection=order_selection)
#    elif instrument =='PEPSI':
#        return PEPSI_get_calib_data(archive, file_rad, fiber=fiber, order_selection=order_selection)
#    else:
#        raise ValueError("Instrument not supported")
#
#
#def get_input_data(instrument, archive, file_rad, mask, fiber='A', skip_ccf=None, skip_s1d=True, order_selection=None):
#
#    if instrument =='HARPS-N':
#        return HARPN_DRSv3_get_input_data(archive, file_rad, mask, fiber=fiber, skip_ccf=skip_ccf, skip_s1d=skip_s1d, order_selection=order_selection)
#    elif instrument =='HARPS':
#        return HARPS_DRSv3_get_input_data(archive, file_rad, mask, fiber=fiber, skip_ccf=skip_ccf, skip_s1d=skip_s1d, order_selection=order_selection)
#    elif instrument =='PEPSI':
#        return PEPSI_get_input_data(archive, file_rad, mask, fiber=fiber, skip_ccf=skip_ccf, skip_s1d=skip_s1d, order_selection=order_selection)
#    else:
#        raise ValueError("Instrument not supported")
